import os

import jax.numpy as jnp
from numpyro.distributions import constraints

from scarlet2 import *
from utils import import_scarlet_test_data

import_scarlet_test_data()
from scarlet_test_data import data_path


def test_quickstart():
    file = jnp.load(os.path.join(data_path, "hsc_cosmos_35.npz"))
    data = jnp.asarray(file["images"])
    centers = [(src['y'], src['x']) for src in file["catalog"]]  # Note: y/x convention!
    weights = jnp.asarray(1 / file["variance"])
    psf = jnp.asarray(file["psfs"])

    frame_psf = GaussianPSF(0.7)
    model_frame = Frame(Box(data.shape), psf=frame_psf)
    obs = Observation(data, weights, psf=ArrayPSF(jnp.asarray(psf))).match(model_frame)

    from functools import partial
    spec_step = partial(relative_step, factor=0.05)

    with Scene(model_frame) as scene:

        for center in centers:
            center = jnp.array(center)
            try:
                spectrum, morph = init.from_gaussian_moments(obs, center, min_corr=0.99)
            except ValueError:
                spectrum = init.pixel_spectrum(obs, center)
                morph = init.compact_morphology()

            Source(center, spectrum, morph)

    # fitting
    parameters = scene.make_parameters()
    for i in range(len(scene.sources)):
        parameters += Parameter(scene.sources[i].spectrum.data, name=f"spectrum:{i}", constraint=constraints.positive,
                                stepsize=spec_step)
        parameters += Parameter(scene.sources[i].morphology.data, name=f"morph:{i}", constraint=constraints.positive,
                                stepsize=0.1)

    maxiter = 200
    scene.set_spectra_to_match(obs, parameters)
    scene_ = scene.fit(obs, parameters, max_iter=maxiter, progress_bar=False)

    # plotting
    norm = plot.AsinhAutomaticNorm(obs)
    plot.scene(scene_, obs, norm=norm, show_model=True, show_rendered=True, show_observed=True, show_residual=True);

    # sampling
    import numpyro.distributions as dist
    from numpyro.infer.initialization import init_to_sample
    parameters = scene_.make_parameters()
    p = scene_.sources[0].spectrum.data
    prior = dist.Normal(p, scale=1)
    parameters += Parameter(p, name=f"spectrum:0", prior=prior)
    mcmc = scene_.sample(obs, parameters, num_samples=200, dense_mass=True, init_strategy=init_to_sample,
                         progress_bar=False)

if __name__ == "__main__":
    test_quickstart()



# from https://github.com/pmelchior/scarlet/blob/master/scarlet/wavelet.py

import jax.numpy as jnp

class Starlet:
    """ A class used to create the Wavelet transform of a cube of images from the 'a trou' algorithm.

        The transform is performed by convolving the image by a seed starlet: the transform of an all-zero
        image with its central pixel set to one. This requires 2-fold padding of the image and an odd pad
        shape. The fft of the seed starlet is cached so that it can be reused in the transform of other
        images that have the same shape.
    """

    def __init__(self, image, coefficients, generation, convolve2D):
        """ Initialise the Starlet object

        Parameters
        ----------
        image: numpy ndarray
            Image in real space.
        coefficients: array
            Starlet transform of the image.
        generation: int
            The generation of the starlet transform (either `1` or `2`).
        convolve2D: array-like
            The filter used to convolve the image and create the wavelets.
            When `convolve2D` is `None` this uses a cubic bspline.
        """
        self._image = image
        self._coeffs = coefficients
        self._generation = generation
        self._convolve2D = convolve2D
        self._norm = None

    @property
    def image(self):
        """The real space image"""
        return self._image

    @property
    def coefficients(self):
        """Starlet coefficients"""
        return self._coeffs


    @staticmethod
    def from_image(image, scales=None, generation=2, convolve2D=None):
        """Generate a set of starlet coefficients for an image

        Parameters
        ----------
        image: array-like
            The image that is converted into starlet coefficients
        scales: int
            The number of starlet scales to use.
            If `scales` is `None` then the maximum number of scales is used.
            Note: this is the length of the coefficients-1, as in the notation
            of `Starck et al. 2011`.
        generation: int
            The generation of the starlet transform (either `1` or `2`).
        convolve2D: array-like
            The filter used to convolve the image and create the wavelets.
            When `convolve2D` is `None` this uses a cubic bspline.

        Returns
        -------
        result: Starlet
            The resulting `Starlet` that contains the image, starlet coefficients,
            as well as the parameters used to generate the coefficients.
        """
        if scales is None:
            scales = get_scales(image.shape)
        coefficients = starlet_transform(image, scales, generation, convolve2D)
        return Starlet(image, coefficients, generation, convolve2D)
    
def bspline_convolve(image, scale):
    """Convolve an image with a bpsline at a given scale.

    This uses the spline
    `h1D = jnp.array([1.0 / 16, 1.0 / 4, 3.0 / 8, 1.0 / 4, 1.0 / 16])`
    from Starck et al. 2011.

    Parameters
    ----------
    image: 2D array
        The image or wavelet coefficients to convolve.
    scale: int
        The wavelet scale for the convolution. This sets the
        spacing between adjacent pixels with the spline.

    """
    # Filter for the scarlet transform. Here bspline
    h1D = jnp.array([1.0 / 16, 1.0 / 4, 3.0 / 8, 1.0 / 4, 1.0 / 16])
    j = scale

    slice0 = slice(None, -2**(j+1))
    slice1 = slice(None, -2**j)
    slice3 = slice(2**j, None)
    slice4 = slice(2**(j+1), None)

    # row
    col = image * h1D[2]
    col = col.at[slice4].add(image[slice0] * h1D[0])
    col = col.at[slice3].add(image[slice1] * h1D[1])
    col = col.at[slice1].add(image[slice3] * h1D[3])
    col = col.at[slice0].add(image[slice4] * h1D[4])

    # column
    result = col * h1D[2]
    result = result.at[:, slice4].add(col[:, slice0] * h1D[0])
    result = result.at[:, slice3].add(col[:, slice1] * h1D[1])
    result = result.at[:, slice1].add(col[:, slice3] * h1D[3])
    result = result.at[:, slice0].add(col[:, slice4] * h1D[4])
    return result


def starlet_transform(image, scales=None, generation=2, convolve2D=None):
    """Perform a scarlet transform, or 2nd gen starlet transform.

    Parameters
    ----------
    image: 2D array
        The image to transform into starlet coefficients.
    scales: int
        The number of scale to transform with starlets.
        The total dimension of the starlet will have
        `scales+1` dimensions, since it will also hold
        the image at all scales higher than `scales`.
    generation: int
        The generation of the transform.
        This must be `1` or `2`.
    convolve2D: function
        The filter function to use to convolve the image
        with starlets in 2D.

    Returns
    -------
    starlet: array with dimension (scales+1, Ny, Nx)
        The starlet dictionary for the input `image`.
    """
    assert len(image.shape) == 2, f"Image should be 2D, got {len(image.shape)}"
    assert generation in (1, 2), f"generation should be 1 or 2, got {generation}"

    scales = get_scales(image.shape, scales)
    c = image
    if convolve2D is None:
        convolve2D = bspline_convolve

    ## wavelet set of coefficients.
    starlet = jnp.zeros((scales + 1,) + image.shape)
    for j in range(scales):
        gen1 = convolve2D(c, j)

        if generation == 2:
            gen2 = convolve2D(gen1, j)
            starlet = starlet.at[j].set(c - gen2)
        else:
            starlet = starlet.at[j].set(c - gen1)

        c = gen1

    starlet = starlet.at[-1].set(c)
    return starlet


def starlet_reconstruction(starlets, generation=2, convolve2D=None):
    """Reconstruct an image from a dictionary of starlets

    Parameters
    ----------
    starlets: array with dimension (scales+1, Ny, Nx)
        The starlet dictionary used to reconstruct the image.
    convolve2D: function
        The filter function to use to convolve the image
        with starlets in 2D.

    Returns
    -------
    image: 2D array
        The image reconstructed from the input `starlet`.
    """
    if generation == 1:
        return jnp.sum(starlets, axis=0)
    if convolve2D is None:
        convolve2D = bspline_convolve
    scales = len(starlets) - 1

    c = starlets[-1]
    for i in range(1, scales + 1):
        j = scales - i
        cj = convolve2D(c, j)
        c = cj + starlets[j]
    return c


def get_scales(image_shape, scales=None):
    """Get the number of scales to use in the starlet transform.

    Parameters
    ----------
    image_shape: tuple
        The 2D shape of the image that is being transformed
    scales: int
        The number of scale to transform with starlets.
        The total dimension of the starlet will have
        `scales+1` dimensions, since it will also hold
        the image at all scales higher than `scales`.
    """
    # Number of levels for the Starlet decomposition
    max_scale = int(jnp.log2(min(image_shape[-2:]))) - 1
    if (scales is None) or scales > max_scale:
        scales = max_scale
    return int(scales)


import astropy.units as u
import astropy.wcs.wcs
import equinox as eqx
import jax.numpy as jnp
from astropy.coordinates import SkyCoord

from .bbox import Box
from .psf import PSF

class Frame(eqx.Module):
    bbox: Box
    psf: PSF = None
    wcs: astropy.wcs.wcs = None
    channels: list

    def __init__(self, bbox, psf=None, wcs=None, channels=None):
        self.bbox = bbox
        self.psf = psf
        self.wcs = wcs
        if channels is None:
            channels = list(range(bbox.shape[0]))
        assert len(channels) == bbox.shape[0]

        self.channels = channels

    def __hash__(self):
        return hash(self.bbox)

    @property
    def C(self):
        return len(self.channels)

    @property
    def pixel_size(self):
        if self.wcs is not None:
            # return get_pixel_size(get_affine(self.wcs)) * 60 * 60  # in arcsec
            return get_scale(self.wcs).mean() * 60 * 60 # in arcsec
        else:
            return 1

    def get_pixel(self, pos):
        """Get the sky coordinates from a world coordinate

        Parameters
        ----------
        pos: jnp.ndarray or SkyCoord
            Coordinates on the sky

        Returns
        ---------
        pixel coordinates in the model frame
        """
        if isinstance(pos, SkyCoord):
            assert self.wcs is not None, "SkyCoord can only be converted with valid WCS"
            wcs_ = self.wcs.celestial  # only use celestial portion
            pixel = jnp.asarray(pos.to_pixel(wcs_), dtype="float32").T
            return pixel
        return pos

    def get_sky_coord(self, pos):
        """Get the sky coordinate from a pixel coordinate

        Parameters
        ----------
        pos: jnp.ndarray
            Coordinates in the pixel space

        Returns
        ----------
        astropy.coordinates.SkyCoord if WCS is set, otherwise pos
        """
        if self.wcs is not None:
            pixels = pos.reshape(-1, 2)
            wcs = self.wcs.celestial  # only use celestial portion
            sky_coord = SkyCoord.from_pixel(pixels[:, 0], pixels[:, 1], wcs)
            return sky_coord
        return pos
    
    def convert_pixel_to(self, target, pixel=None):
        """Converts pixel coordinates from this frame to `target` Frame

            Parameters
            ----------
            target: `~scarlet2.Frame`
                target frame
            pixel: `array`, pixel coordinates in this frame
                If not set, convert all pixels in this frame

            Returns
            -------
            coord_target: `array`
                coordinates at the location of `coord` in the target frame
        """

        if pixel is None:
            y, x = jnp.indices(self.bbox.shape[-2:], dtype="float32")
            pixel = jnp.stack((y.flatten(), x.flatten()), axis=1)

        ra_dec = self.get_sky_coord(pixel)
        return target.get_pixel(ra_dec)
    
    def u_to_pixel(self, size):
        """Converts a size un astropy.units.Quantity to pixel size according 
           to this frame WCS

            Parameters
            ----------
            size: `astropy.units.Quantity`, must be PhysicalType("angle")

            Returns
            -------
            size in pixels
        """
        assert u.get_physical_type(size) == "angle"

        # first computer the pixel size
        pixel_size = get_pixel_size(
            get_affine(self.wcs.celestial) # only use celestial portion
        ) * 60 * 60 # in arcsec/pixel
        
        return size.to(u.arcsec).value / pixel_size

    @staticmethod
    def from_observations(
        observations, model_psf=None, model_wcs=None, obs_id=None, coverage="union"
    ):
        """Generates a suitable model frame for a set of observations.

        This method generates a frame from a set of observations by identifying the highest resolution
        and the smallest PSF and use them to construct a common frame for all observations.

        Parameters
        ----------
        observations: array of `scarlet.Observation` objects
            array that contains Observations to match onto a common frame
        model_psf: `scarlet.PSF`
            PSF of the model frame, to which all observations are to be deconvolved.
            If None, uses the smallest PSF across all observations and channels.
        model_wcs: `astropy.wcs.WCS`
            WCS for the model frame. If None, uses transformation of the observation
            with the smallest pixels.
        obs_id: int
            index of the reference observation
            If set to None, uses the observation with the smallest pixels.
        coverage: "union" or "intersection"
            Sets the frame to incorporate the pixels covered by any observation ('union')
            or by all observations ('intersection').
        """
        assert coverage in ["union", "intersection"]

        if not hasattr(observations, "__iter__"):
            observations = (observations,)

        # Array of pixel sizes for each observation
        pix_tab = []
        # Array of psf size for each psf of each observation
        fat_psf_size = None
        small_psf_size = None
        channels = []
        # Create frame channels and find smallest and largest psf
        for c, obs in enumerate(observations):

            # Concatenate all channels
            channels = channels + obs.frame.channels

            # concatenate all pixel sizes
            h_temp = get_pixel_size(get_affine(obs.frame.wcs))

            pix_tab.append(h_temp)
            # Looking for the sharpest and the fatest psf
            psf = obs.frame.psf.morphology
            for psf in psf:
                psf_size = get_psf_size(psf) * h_temp
                if (fat_psf_size is None) or (psf_size > fat_psf_size):
                    fat_psf_size = psf_size
                if (obs_id is None) or (c == obs_id):
                    if (model_psf is None) and (
                        (small_psf_size is None) or (psf_size < small_psf_size)
                    ):
                        small_psf_size = psf_size
                        model_psf_temp = ImagePSF(psf[np.newaxis, :, :])
                        psf_h = h_temp

        # Find a reference observation. Either provided by obs_id or as the observation with the smallest pixel
        if obs_id is None:
            p = jnp.array(pix_tab)
            obs_ref = observations[jnp.where(p == p.min())[0][0]]
        else:
            # Frame defined from obs_id
            obs_ref = observations[obs_id]

        # Reference wcs
        if model_wcs is None:
            model_wcs = obs_ref.frame.wcs

        # Scale of the smallest pixel
        h = get_pixel_size(get_affine(model_wcs))

        # TODO:
        # # If needed and psf is not provided: interpolate psf to smallest pixel
        # if model_psf is None:
        #     # If the reference PSF is not at the highest pixel resolution, make it!
        #     if psf_h > h:
        #         angle, h = interpolation.get_angles(model_wcs, obs.wcs)
        #         model_psf = PSF(
        #             interpolation.sinc_interp_inplace(model_psf_temp, psf_h, h, angle)
        #         )
        #     else:
        #         model_psf = model_psf_temp

        # Dummy frame for WCS computations
        model_shape = (len(channels), 0, 0)

        model_frame = Frame(
            Box(model_shape), channels=channels, psf=model_psf, wcs=model_wcs
        )

        # Determine overlap of all observations in pixel coordinates of the model frame
        for c, obs in enumerate(observations):

            if model_frame.wcs is obs.frame.wcs:
                this_box = obs_ref.frame.bbox[-2:]
            else:
                obs_coord = obs.frame.convert_pixel_to(model_frame)
                y_min = jnp.floor(jnp.min(obs_coord[:, 0])).astype("int")
                x_min = jnp.floor(jnp.min(obs_coord[:, 1])).astype("int")
                y_max = jnp.ceil(jnp.max(obs_coord[:, 0])).astype("int")
                x_max = jnp.ceil(jnp.max(obs_coord[:, 1])).astype("int")
                this_box = Box.from_bounds((y_min, y_max + 1), (x_min, x_max + 1))

            if c == 0:
                model_box = this_box
            else:
                if coverage == "union":
                    model_box |= this_box
                else:
                    model_box &= this_box

        # pad by the size of the widest psf to prevent leakage across the frame edge
        ny, nx = model_box.shape
        pad_size = fat_psf_size / h / 2
        offset = (jnp.round(pad_size).astype("int"), jnp.round(pad_size).astype("int"))
        model_box -= offset

        model_box_shape = tuple(int(s + 2 * o) for s, o in zip(model_box.shape, offset))

        # move the reference pixel of the model wcs to the 0/0 pixel of the new shape
        model_wcs = model_wcs.deepcopy()
        model_wcs.wcs.crpix -= model_box.origin
        model_wcs.array_shape = model_box.shape

        # recreate the model frame with the correct shape
        # frame_shape = (len(channels), model_box_shape)
        frame_shape = (len(channels),) + model_box_shape

        model_frame = Frame(
            Box(frame_shape), channels=channels, psf=model_psf, wcs=model_wcs
        )

        # Match observations to this frame
        for obs in observations:
            obs.match(model_frame)

        return model_frame


def get_psf_size(psf):
    """Measures the size of a psf by computing the size of the area in 3 sigma around the center.

    This is an approximate method to estimate the size of the psf for setting the size of the frame,
    which does not require a precise measurement.

    Parameters
    ----------
        PSF: `scarlet.PSF` object
            PSF for whic to compute the size
    Returns
    -------
        sigma3: `float`
            radius of the area inside 3 sigma around the center in pixels
    """
    # Normalisation by maximum
    psf_frame = psf / jnp.max(psf)

    # Pixels in the FWHM set to one, others to 0:
    psf_frame = jnp.where(psf_frame > 0.5, 1., 0.)

    # Area in the FWHM:
    area = jnp.sum(psf_frame)

    # Diameter of this area
    d = 2 * (area / jnp.pi) ** 0.5

    # 3-sigma:
    sigma3 = 3 * d / (2 * (2 * jnp.log(2)) ** 0.5)

    return sigma3

def get_affine(wcs):
    try:
        model_affine = wcs.wcs.pc
    except AttributeError:
        try:
            model_affine = wcs.cd
        except AttributeError:
            model_affine = wcs.wcs.cd
    return model_affine


def get_pixel_size(model_affine):
    """Extracts the pixel size from a wcs, and returns it in deg/pixel"""
    pix = jnp.sqrt(
        jnp.abs(model_affine[0, 0])
        * jnp.abs(model_affine[1, 1] - model_affine[0, 1] * model_affine[1, 0])
    )
    return pix

def get_scale(wcs):
    """
    Return WCS axis scales in deg/pixel
    """
    model_affine = get_affine(wcs)
    c1 = (model_affine[0,:2]**2).sum()**0.5
    c2 = (model_affine[1,:2]**2).sum()**0.5
    return jnp.array([c1, c2])
    
def get_angle(wcs):
    """
    Return WCS rotation angle in rad
    """
    model_affine = get_affine(wcs)
    c = get_scale(wcs)
    c = c.reshape([c.shape[-1],1])
    R = model_affine[:2, :2] / c # removing the scaling factors from the pc

    if R[0,0]==0.:
        return jnp.arcsin(R[0,1])
    else:
        return jnp.arctan(R[0,1]/R[0,0])

def get_sign(wcs):
    """
    Return WCS flip signs
    """
    model_affine = get_affine(wcs)
    c = get_scale(wcs)
    c = c.reshape([c.shape[-1],1])
    R = model_affine[:2, :2] / c # removing the absolute scaling factors from the pc

    if R[0,0]==0.:
        phi = jnp.arcsin(R[0,1])
    else:
        phi = jnp.arctan(R[0,1]/R[0,0])
    
    R_inv = jnp.array([[jnp.cos(phi), -jnp.sin(phi)],
                    [jnp.sin(phi), jnp.cos(phi)]])
    
    R = R_inv @ R
    return jnp.round(jnp.diag(R))


import equinox as eqx
import jax
import jax.numpy as jnp

from . import Scenery
from .bbox import overlap_slices
from .frame import Frame
from .module import Module, Parameters
from .renderer import ChannelRenderer
from .spectrum import ArraySpectrum


class Scene(Module):
    frame: Frame = eqx.field(static=True)
    sources: list

    def __init__(self, frame):
        self.frame = frame
        self.sources = list()

    def __call__(self):
        model = jnp.zeros(self.frame.bbox.shape)
        for source in self.sources:
            model += self._eval_src_in_frame(source)
        return model

    def _eval_src_in_frame(self, source):
        model_ = source()
        # cut out region from model, add single source model
        bbox, bbox_ = overlap_slices(self.frame.bbox, source.bbox, return_boxes=True)
        sub_model_ = jax.lax.dynamic_slice(model_, bbox_.start, bbox_.shape)

        # add model_ back in full model
        model = jnp.zeros(self.frame.bbox.shape)
        model = jax.lax.dynamic_update_slice(model, sub_model_, bbox.start)
        return model

    def __enter__(self):
        # context manager to register sources
        # purpose is to provide scene.frame to source inits that will need some of its information
        # also allows us to append the sources automatically to the scene
        Scenery.scene = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        Scenery.scene = None

    def sample(self, observations, parameters, seed=0, num_warmup=100, num_samples=200, progress_bar=True, **kwargs):
        # uses numpyro NUTS on all non-fixed parameters
        # requires that those have priors set
        try:
            import numpyro
            import numpyro.distributions as dist
            import numpyro.distributions.constraints as constraints
            from numpyro.infer import MCMC, NUTS
        except ImportError:
            raise ImportError("scarlet2.Scene.sample() requires numpyro.")

        # making sure we can iterate
        if not isinstance(observations, (list, tuple)):
            observations = (observations,)

        # helper class to turn observation likelihood(s) into numpyro distribution
        class ObsDistribution(dist.Distribution):
            support = constraints.real_vector

            def __init__(self, obs, model, validate_args=None):
                self.obs = obs
                self.model = model
                event_shape = jnp.shape(model)
                super().__init__(
                    event_shape=event_shape,
                    validate_args=validate_args,
                )

            def sample(self, key, sample_shape=()):
                raise NotImplementedError

            def mean(self):
                return self.obs.render(self.model)

            @dist.util.validate_sample
            def log_prob(self, value):
                # numpyro needs sampling distribution of data (=value), not likelihood function of parameters
                return self.obs._log_likelihood(self.model, value)

        # find all non-fixed parameters and their priors
        priors = {p.name: p.prior for p in parameters}
        has_none = any(prior is None for prior in priors.values())
        if has_none:
            from pprint import pformat
            msg = f"All parameters need to have priors set. Got:\n{pformat(priors)}"
            raise AttributeError(msg)

        # define the pyro model, where every parameter becomes a sample,
        # and the observations sample from their likelihood given the rendered model
        def pyro_model(model):
            samples = tuple(numpyro.sample(p.name, p.prior) for p in parameters)
            model_ = model.replace(parameters, samples)
            pred = model_()  # create prediction once for all observations
            # dealing with multiple observations
            for i, obs_ in enumerate(observations):
                numpyro.sample(f"obs.{i}", ObsDistribution(obs_, pred), obs=obs_.data)

        from numpyro.infer import MCMC, NUTS
        nuts_kernel = NUTS(pyro_model, **kwargs)
        mcmc = MCMC(nuts_kernel, num_warmup=num_warmup, num_samples=num_samples, progress_bar=progress_bar)
        rng_key = jax.random.PRNGKey(seed)
        mcmc.run(rng_key, self)
        return mcmc

    def fit(self, observations, parameters, schedule=None, max_iter=100, e_rel=1e-4, progress_bar=True, callback=None,
            **kwargs):
        # optax fit with adam optimizer
        # Transforms constrained parameters into unconstrained ones
        # and filters out fixed parameters
        # TODO: check alternative optimizers
        try:
            import tqdm
            import optax
            import optax._src.base as base
            from numpyro.distributions.transforms import biject_to
        except ImportError:
            raise ImportError("scarlet2.Scene.fit() requires optax and numpyro.")

        # making sure we can iterate
        if not isinstance(observations, (list, tuple)):
            observations = (observations,)
        assert isinstance(parameters, Parameters)

        # make a stepsize tree
        where = lambda model: model.get(parameters)
        replace = tuple(p.stepsize for p in parameters)
        steps = eqx.tree_at(where, self, replace=replace)

        def scale_by_stepsize() -> base.GradientTransformation:
            # adapted from optax.scale_by_param_block_norm()
            def init_fn(params):
                del params
                return base.EmptyState()

            def update_fn(updates, state, params):
                if params is None:
                    raise ValueError(base.NO_PARAMS_MSG)
                updates = jax.tree_util.tree_map(
                    # lambda u, step, param: -step * u if not callable(step) else -step(param,niter) * u,
                    lambda u, s, p: -s * u if not callable(s) else -s(p) * u,
                    # minus because we want gradient descent
                    updates, steps, params)
                return updates, state

            return base.GradientTransformation(init_fn, update_fn)

        # run adam, followed by stepsize adjustments
        optim = optax.chain(
            optax.scale_by_adam(**kwargs),
            optax.scale_by_schedule(schedule if callable(schedule) else lambda x: 1),
            scale_by_stepsize(),
        )

        # transform to unconstrained parameters
        scene = _constraint_replace(self, parameters, inv=True)

        # get optimizer initialized with the optimization parameters
        filter_spec = self.get_filter_spec(parameters)
        if filter_spec is None:
            opt_state = optim.init(scene)
        else:
            opt_state = optim.init(eqx.filter(scene, filter_spec))

        with tqdm.trange(max_iter, disable=not progress_bar) as t:
            for step in t:
                # optimizer step
                scene, loss, opt_state, convergence = _make_step(scene, observations, parameters, optim, opt_state,
                                                                 filter_spec=filter_spec)

                # compute max change across all non-fixed parameters for convergence test
                max_change = jax.tree_util.tree_reduce(lambda a, b: max(a, b), convergence)

                # report current iteration results to callback
                if callback is not None:
                    scene_ = _constraint_replace(scene, parameters)
                    callback(scene_, convergence, loss)

                # Log the loss and max_change in the tqdm progress bar
                t.set_postfix(loss=f"{loss:08.2f}", max_change=f"{max_change:1.6f}")

                # test convergence
                if max_change < e_rel:
                    break

        return _constraint_replace(scene, parameters)  # transform back to constrained variables

    def set_spectra_to_match(self, observations, parameters):
        """Sets the spectra of every source in the scene to match the observations.

        Computes the best-fit amplitude of the rendered model of all components in every
        channel of every observation as a linear inverse problem.

        Parameters
        ----------
        observations: `scarlet2.Observation` or list thereof
        """

        if not hasattr(observations, "__iter__"):
            observations = (observations,)

        # extract multi-channel model for every source
        spectrum_parameters = []
        models = []
        for i, src in enumerate(self.sources):
            if isinstance(src.spectrum, ArraySpectrum):
                # search for spectrum.data in parameters
                for p in parameters:
                    if p.node is src.spectrum.data:
                        spectrum_parameters.append(i)
                        # update source to have flat spectrum
                        src = eqx.tree_at(lambda src: src.spectrum.data, src, jnp.ones_like(p.node))
                        break

            # evaluate the model for any source so that fit includes it even if its spectrum is not updated
            model = self._eval_src_in_frame(src)  # assumes all sources are single components

            # check for models with identical initializations, see scarlet repo issue #282
            # if duplicate: raise ValueError
            for l in range(len(models)):
                if jnp.allclose(model, models[l]):
                    message = f"Source {i} has a model identical to source {l}.\n"
                    message += "This is likely not intended, and the second source should be deleted."
                    raise ValueError(message)
            models.append(model)

        models = jnp.array(models)
        K = len(models)

        for obs in observations:
            # independent channels, no mixing
            # solve the linear inverse problem of the amplitudes in every channel
            # given all the rendered morphologies
            # spectrum = (M^T Sigma^-1 M)^-1 M^T Sigma^-1 * im
            C = obs.frame.C
            images = obs.data
            weights = obs.weights
            morphs = jnp.stack([obs.render(model) for model in models], axis=0)
            spectra = jnp.zeros((K, C))
            for c in range(C):
                im = images[c].reshape(-1)
                w = weights[c].reshape(-1)
                m = morphs[:, c, :, :].reshape(K, -1)
                mw = m * w[None, :]
                # check if all components have nonzero flux in c.
                # because of convolutions, flux can be outside the box,
                # so we need to compare weighted flux with unweighted flux,
                # which is the same (up to a constant) for constant weights.
                # so we check if *most* of the flux is from pixels with non-zero weight
                nonzero = jnp.sum(mw, axis=1) / jnp.sum(m, axis=1) / jnp.mean(w) > 0.1
                nonzero = jnp.flatnonzero(nonzero)
                if len(nonzero) == K:
                    covar = jnp.linalg.inv(mw @ m.T)
                    spectra = spectra.at[:, c].set(covar @ m @ (im * w))
                else:
                    covar = jnp.linalg.inv(mw[nonzero] @ m[nonzero].T)
                    spectra = spectra.at[nonzero, c].set(covar @ m[nonzero] @ (im * w))

            # update the parameters with the best-fit spectrum solution
            channel_map = ChannelRenderer(self.frame, obs.frame).channel_map
            noise_bg = 1 / jnp.median(jnp.sqrt(obs.weights), axis=(-2, -1))
            for i in spectrum_parameters:
                src_ = self.sources[i]
                # faint galaxy can have erratic solution, bound from below by noise_bg
                v = src_.spectrum.data.at[channel_map].set(jnp.maximum(spectra[i], noise_bg))
                self.sources[i] = eqx.tree_at(lambda src: src.spectrum.data, src_, v)

def _constraint_replace(self, parameters, inv=False):
    # replace any parameter with constraint into unconstrained ones by calling its constraint bijector
    # return transformed pytree
    where_in = lambda model: model.get(parameters)
    param_values = where_in(self)
    if not inv:
        replace = tuple(
            p.constraint_transform(v) if p.constraint is not None else v for p, v in zip(parameters, param_values))
    else:
        replace = tuple(
            p.constraint_transform.inv(v) if p.constraint is not None else v for p, v in zip(parameters, param_values))

    return eqx.tree_at(where_in, self, replace=replace)

# update step for optax optimizer
@eqx.filter_jit
def _make_step(model, observations, parameters, optim, opt_state, filter_spec=None):
    def loss_fn(model):
        if any(param.constraint is not None for param in parameters):
            # parameters now obey constraints
            # transformation happens in the grad path, so gradients are wrt to unconstrained variables
            # likelihood and prior grads transparently apply the Jacobians of these transformations
            model = _constraint_replace(model, parameters)

        pred = model()
        log_like = sum(obs.log_likelihood(pred) for obs in observations)

        param_values = model.get(parameters)
        log_prior = sum(param.prior.log_prob(value)
                        for param, value in zip(parameters, param_values)
                        if param.prior is not None
                        )
        return -(log_like + log_prior)

    if filter_spec is None:
        loss, grads = eqx.filter_value_and_grad(loss_fn)(model)
    else:
        @eqx.filter_value_and_grad
        def filtered_loss_fn(diff_model, static_model):
            model = eqx.combine(diff_model, static_model)
            return loss_fn(model)

        diff_model, static_model = eqx.partition(model, filter_spec)
        loss, grads = filtered_loss_fn(diff_model, static_model)

    updates, opt_state = optim.update(grads, opt_state, model)
    model_ = eqx.apply_updates(model, updates)

    # for convergence criterion: compute norms of parameters and updates
    norm = lambda x, dx: 0 if dx is None else jnp.linalg.norm(dx) / jnp.linalg.norm(x)
    convergence = jax.tree_util.tree_map(lambda x, dx: norm(x, dx), *(model, updates))

    return model_, loss, opt_state, convergence



import astropy.units as u
import equinox as eqx
import jax.numpy as jnp
import jax.scipy

from . import Scenery
from . import measure
from .module import Module
from .wavelets import starlet_transform, starlet_reconstruction


class Morphology(Module):

    @property
    def shape(self):
        raise NotImplementedError

    def normalize(self, x):
        return x / x.max()


class ArrayMorphology(Morphology):
    data: jnp.array

    def __init__(self, data):
        self.data = data

    def __call__(self, **kwargs):
        return self.normalize(self.data)

    @property
    def shape(self):
        return self.data.shape


class ProfileMorphology(Morphology):
    size: float
    ellipticity: (None, jnp.array)
    _shape: tuple = eqx.field(static=True, init=False, repr=False)

    def __init__(self, size, ellipticity=None, shape=None):

        if isinstance(size, u.Quantity):
            try:
                size = Scenery.scene.frame.u_to_pixel(size)
            except AttributeError:
                print("`size` defined in astropy units can only be used within the context of a Scene")
                print("Use 'with Scene(frame) as scene: (...)'")
                raise

        self.size = size
        self.ellipticity = ellipticity

        # default shape: square 10x size
        if shape is None:
            # explicit call to int() to avoid bbox sizes being jax-traced
            size = int(jnp.ceil(10 * self.size))
            # odd shapes for unique center pixel
            if size % 2 == 0:
                size += 1
            shape = (size, size)
        self._shape = shape

    @property
    def shape(self):
        return self._shape

    def f(self, R2):
        raise NotImplementedError

    def __call__(self, delta_center=jnp.zeros(2)):

        _Y = jnp.arange(-(self.shape[-2] // 2), self.shape[-2] // 2 + 1, dtype=float) + delta_center[-2]
        _X = jnp.arange(-(self.shape[-1] // 2), self.shape[-1] // 2 + 1, dtype=float) + delta_center[-1]

        if self.ellipticity is None:
            R2 = _Y[:, None] ** 2 + _X[None, :] ** 2
        else:
            e1, e2 = self.ellipticity
            g_factor = 1 / (1. + jnp.sqrt(1. - (e1 ** 2 + e2 ** 2)))
            g1, g2 = self.ellipticity * g_factor
            __X = ((1 - g1) * _X[None, :] - g2 * _Y[:, None]) / jnp.sqrt(
                1 - (g1 ** 2 + g2 ** 2)
            )
            __Y = (-g2 * _X[None, :] + (1 + g1) * _Y[:, None]) / jnp.sqrt(
                1 - (g1 ** 2 + g2 ** 2)
            )
            R2 = __Y ** 2 + __X ** 2

        R2 /= self.size ** 2
        R2 = jnp.maximum(R2, 1e-3)  # prevents infs at R2 = 0
        morph = self.normalize(self.f(R2))
        return morph


class GaussianMorphology(ProfileMorphology):

    def f(self, R2):
        return jnp.exp(-R2 / 2)

    def __call__(self, delta_center=jnp.zeros(2)):

        # faster circular 2D Gaussian: instead of N^2 evaluations, use outer product of 2 1D Gaussian evals
        if self.ellipticity is None:
            _Y = jnp.arange(-(self.shape[-2] // 2), self.shape[-2] // 2 + 1, dtype=float) + delta_center[-2]
            _X = jnp.arange(-(self.shape[-1] // 2), self.shape[-1] // 2 + 1, dtype=float) + delta_center[-1]

            # with pixel integration
            f = lambda x, s: 0.5 * (
                    1 - jax.scipy.special.erfc((0.5 - x) / jnp.sqrt(2) / s) +
                    1 - jax.scipy.special.erfc((0.5 + x) / jnp.sqrt(2) / s)
            )
            # # without pixel integration
            # f = lambda x, s: jnp.exp(-(x ** 2) / (2 * s ** 2)) / (jnp.sqrt(2 * jnp.pi) * s)

            return self.normalize(jnp.outer(f(_Y, self.size), f(_X, self.size)))

        else:
            return super().__call__(delta_center)

    @staticmethod
    def from_image(image):
        assert image.ndim == 2
        center = measure.centroid(image)
        # compute moments and create Gaussian from it
        g = measure.moments(image, center=center, N=2)
        return GaussianMorphology.from_moments(g)

    @staticmethod
    def from_moments(g):
        T = g.size
        ellipticity = g.ellipticity

        # create image of Gaussian with these 2nd moments
        if jnp.isfinite(T) and jnp.isfinite(ellipticity).all():
            morph = GaussianMorphology(T, ellipticity)
        else:
            raise ValueError(
                f"Gaussian morphology not possible with size={T}, and ellipticity={ellipticity}!")
        return morph


class SersicMorphology(ProfileMorphology):
    n: float

    def __init__(self, n, size, ellipticity=None):
        self.n = n
        super().__init__(size, ellipticity=ellipticity)

    def f(self, R2):
        n = self.n
        n2 = n * n
        # simplest form of bn: Capaccioli (1989)
        # bn = 1.9992 * n - 0.3271
        #
        # better treatment in  Ciotti & Bertin (1999), eq. 18
        # stable to n > 0.36, with errors < 10^5
        bn = 2 * n - 0.333333 + 0.009877 / n + 0.001803 / n2 + 0.000114 / (n2 * n) - 0.000072 / (n2 * n2)

        # MacArthur, Courteau, & Holtzman (2003), eq. A2
        # much more stable for n < 0.36
        # not using it here to avoid if clause in jitted code
        #     bn = 0.01945 - 0.8902 * n + 10.95 * n2 - 19.67 * n2 * n + 13.43 * n2 * n2

        # Graham & Driver 2005, eq. 1
        # we're given R^2, so we use R2^(0.5/n) instead of 1/n
        return jnp.exp(-bn * (R2 ** (0.5 / n) - 1))


prox_plus = lambda x: jnp.maximum(x, 0)
prox_soft = lambda x, thresh: jnp.sign(x) * prox_plus(jnp.abs(x) - thresh)
prox_soft_plus = lambda x, thresh: prox_plus(prox_soft(x, thresh))


class StarletMorphology(Morphology):
    coeffs: jnp.ndarray
    l1_thresh: float = eqx.field(default=1e-2, static=True)
    positive: bool = eqx.field(default=True, static=True)

    def __call__(self, **kwargs):
        if self.positive:
            f = prox_soft_plus
        else:
            f = prox_soft
        return self.normalize(starlet_reconstruction(f(self.coeffs, self.l1_thresh)))

    @property
    def shape(self):
        return self.coeffs.shape[-2:]  # wavelet coeffs: scales x n1 x n2

    @staticmethod
    def from_image(image, **kwargs):
        # Starlet transform of image (n1,n2) into coefficient with 3 dimensions: (scales+1,n1,n2)
        coeffs = starlet_transform(image)
        return StarletMorphology(coeffs, **kwargs)



# -------------------------------------------------- #
# This class is used to calculate the gradient       #
# of the log-probability via calling the score prior #
# ScoreNet model. A custom vjp is created to return  #
# the score prior when calling jax.grad()            #
# -------------------------------------------------- #
try:
    import numpyro.distributions as dist
    import numpyro.distributions.constraints as constraints

except ImportError:
    raise ImportError("scarlet2.nn requires numpyro.")

import jax.numpy as jnp
from jax import custom_vjp
from jax import vjp

def pad_fwd(x, model_shape):
    """Zero-pads the input image to the model size
    
    Parameters
    ----------
    x : jnp.array
        data to be padded
    model_shape : tuple
        shape of the prior model to be used
        
    Returns
    -------
    x : jnp.array
        data padded to same size as model_shape
    pad: tuple
        padding amount in every dimension
    """
    assert all(model_shape[d] >= x.shape[d] for d in range(x.ndim)), "Model size must be larger than data size"
    if model_shape == x.shape:
        pad = 0
        return x, pad

    pad = tuple(
        # even padding
        (int(gap / 2), int(gap / 2)) if (gap := model_shape[d] - x.shape[d]) % 2 == 0
        # uneven padding
        else (int(gap // 2), int(gap // 2) + 1)
        # over all dimensions
        for d in range(x.ndim)
    )
    # perform the zero-padding
    x = jnp.pad(x, pad, 'constant', constant_values=0)
    return x, pad
    
    
# reverse pad back to original size
def pad_back(x, pad):
    """Removes the zero-padding from the input image

    Parameters
    ---------
    x : jnp.array
        padded data to same size as model_shape
    pad: tuple
        padding amount in every dimension
        
    Returns
    -------
    x : jnp.array
        data returned to it pre-pad shape
    """
    slices = tuple(slice(low, -hi) if hi > 0 else slice(low, None) for (low, hi) in pad)
    return x[slices]


# calculate score function (jacobian of log-probability)
def calc_grad(x, model):
    """Calculates the gradient of the log-prior 
    using the ScoreNet model chosen
    
    Parameters
    ----------
    x : array of the data
    model : the model to calculate the score function

    Returns
    -------
    score_func : array of the score function
    """
    # cast to float32, expand to (batch, shape), and pad to match the shape of the score model
    x_, pad = pad_fwd(jnp.float32(x), model.shape)

    # run score model, expects (batch, shape)
    if jnp.ndim(x) == len(model.shape):
        x_ = jnp.expand_dims(x_, axis=0)
    score_func = model(x_)
    if jnp.ndim(x) == len(model.shape):
        score_func = jnp.squeeze(score_func, axis=0)

    # remove padding
    if pad != 0:
        score_func = pad_back(score_func, pad)
    return score_func


# jax gradient function to calculate jacobian
def vgrad(f, x):
    y, vjp_fn = vjp(f, x)
    return vjp_fn(jnp.ones(y.shape))[0]

# Here we define a custom vjp for the log_prob function
# such that for gradient calls in jax, the score prior
# is returned
from functools import partial

@partial(custom_vjp, nondiff_argnums=(0,))
def _log_prob(model, x):
    return 0.

def log_prob_fwd(model, x):
    score_func = calc_grad(x, model)
    return 0., score_func  # cannot directly call log_prob in Class object

def log_prob_bwd(model, res, g):
    score_func = res  # Get residuals computed in f_fwd
    return (g * score_func,)  # create the vector (g) jacobian (score_func) product

# register the custom vjp 
_log_prob.defvjp(log_prob_fwd, log_prob_bwd)


# inheritance from Distribution class
class ScorePrior(dist.Distribution):
    support = constraints.real_vector
    """Prior distribution based on a neural network"""

    def __init__(self, model, validate_args=None):
        self.model = model

        super().__init__(
            event_shape=model.shape,
            validate_args=validate_args,
        )

    def sample(self, key, sample_shape=()):
        # TODO: add ability to draw samples from the prior, if desired
        raise NotImplementedError

    def mean(self):
        raise NotImplementedError
    
    def log_prob(self, x):
        return _log_prob(self.model, x)

# define a class for temperature adjustable prior
class TempScore(ScorePrior):
    """Temperature adjustable ScorePrior"""
    def __init__(self, model, temp=0.02):
        self.model = model
        self.temp = temp
    def __call__(self, x):
        return self.model(x, t=self.temp)




import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import varname

from astropy.coordinates import SkyCoord
import astropy.units as u

class Module(eqx.Module):

    def __call__(self):
        raise NotImplementedError

    def make_parameters(self):
        return Parameters(self)

    def get(self, parameters):
        assert isinstance(parameters, Parameters)
        return parameters.extract_from(self)

    def replace(self, parameters, values):
        """Replace attribute with given name by other value
        """
        where = lambda model: model.get(parameters)
        return eqx.tree_at(where, self, replace=values)

    def get_filter_spec(self, parameters):
        """Get equinox filter_spec for all fields named in parameters
        """
        filtered = jax.tree_util.tree_map(lambda _: False, self)
        where = lambda model: model.get(parameters)
        values = (True,) * len(parameters)
        filtered = eqx.tree_at(where, filtered, replace=values)
        if all(jax.tree_util.tree_leaves(filtered)):
            return None
        return filtered


class Parameter:

    def __init__(self, node, name=None, constraint=None, prior=None, stepsize=0):

        if name is None:
            self.name = varname.argname('node', vars_only=False)
        else:
            self.name = name
        self.node = node

        self.constraint = constraint
        
        self.prior = prior
        self.stepsize = stepsize
    
    def set_constraint(self):
        if self.constraint is not None:
            try:
                from numpyro.distributions.transforms import biject_to
            except ImportError:
                raise ImportError("scarlet2.Parameter requires numpyro.")
            # transformation to unconstrained parameters
            self.constraint_transform = biject_to(self.constraint)

            # check if parameter is valid under transform
            unconstrained = self.constraint_transform.inv(self.node)
            if not jnp.isfinite(unconstrained).all():
                raise ValueError(f"Parameter {self.name} has infeasible values for constraint {self.constraint}!")

    def __repr__(self):
        # equinox-like formatting
        chunks = []
        for name in ["name", "node", "constraint", "prior", "stepsize"]:
            field = getattr(self, name)
            if name == "node" and isinstance(field, jax.Array):
                field = eqx._pretty_print._pformat_array(field, short_arrays=True)
            chunks.append(f"  {name}={field}")
        inner = ",\n".join(chunks)
        mess = f"{self.__class__.__name__}(\n{inner}\n)"
        return mess


class Parameters:
    def __init__(self, base):
        self.base = base
        self._base_leaves = jtu.tree_leaves(base)
        self._params = list()
        self._leave_idx = list()

    def __repr__(self):
        # equinox-like formatting
        mess = f"{self.__class__.__name__}(\n"
        mess += f"  base={self.base.__class__.__name__},\n"
        mess += f"  parameters=[\n"
        chunks = []
        for p in self._params:
            mess_ = p.__repr__()
            chunk = ""
            for line in mess_.splitlines(keepends=True):
                chunk += "    " + line
            chunks.append(chunk)

        mess += ",\n".join(chunks)
        mess += "\n  ]\n"
        mess += ")"
        return mess

    def __iadd__(self, parameter):
        assert isinstance(parameter, Parameter)
        found = False
        for i, leaf in enumerate(self._base_leaves):
            if leaf is parameter.node:
                parameter = self.to_pixels(parameter)
                parameter.set_constraint()
                self._params.append(parameter)
                self._leave_idx.append(i)
                found = True
                break
        if not found:
            mess = f"Parameter '{parameter.name}' not in {self.base.__class__.__name__}!"
            raise RuntimeError(mess)
        return self

    def __isub__(self, name):
        for i, param in enumerate(self._params):
            if param.name == name:
                del self._params[i]
                del self._leave_idx[i]
                break
        return self

    def __getitem__(self, i):
        return self._params[i]

    def __len__(self):
        return len(self._params)

    def extract_from(self, root):
        # create function that ingests root and returns all nodes
        assert jtu.tree_structure(root) == jtu.tree_structure(self.base)
        root_leaves = jtu.tree_leaves(root)
        return tuple(root_leaves[idx] for idx in self._leave_idx)
    

    def to_pixels(self, parameter):
        frame = self.base.frame
        used_sky_coords_prior = False

        for fieldname in ['node', 'constraint', 'prior', 'stepsize']:
            field = getattr(parameter, fieldname)
            if isinstance(field, u.Quantity):
                setattr(parameter, fieldname, frame.u_to_pixel(field))
            if isinstance(field, SkyCoord):
                    setattr(parameter, fieldname, frame.get_pixel(field))
            for name in dir(field):
                try:
                    attrib = getattr(field, name)
                    if isinstance(attrib, u.Quantity):
                        setattr(field, name, frame.u_to_pixel(attrib))
                    if isinstance(attrib, SkyCoord):
                        setattr(field, name, frame.get_pixel(attrib))
                        used_sky_coords_prior = (fieldname == 'prior')
                except:
                    pass

            if used_sky_coords_prior:
                try:
                    import numpyro.distributions as dist
                except ImportError:
                    raise ImportError("scarlet2.Parameter requires numpyro.")
                
                # converting SkyCoord to Array in numpyro distributions requires
                # to update batch and event shape 
                batch_shape = max([getattr(field, name).shape 
                                   for name in field.reparametrized_params])
                setattr(field, '_batch_shape', batch_shape)
                setattr(parameter, fieldname, dist.Independent(field, 1))
                
            used_sky_coords_prior = False

        return parameter


def relative_step(x, *args, factor=0.01, minimum=1e-6):
    """Step size set at `factor` times the norm of `x`

    As step size functions have the signature (array, int) -> float, *args captures, 
    and then ignores, the iteration counter.
    """
    return jnp.maximum(minimum, factor * jnp.linalg.norm(x))


import equinox as eqx
import jax.numpy as jnp

class Box(eqx.Module):
    """Bounding Box for an object

    A Bounding box describes the location of a data unit in the global/model coordinate
    system. It is used to identify spatial and channel overlap and to map from model
    to observed frames and back.

    The `BBox` code is agnostic about the meaning of the dimensions.
    We generally use this convention:

    - 2D shapes denote (Height, Width)
    - 3D shapes denote (Channels, Height, Width)
    """

    shape: tuple
    origin: tuple

    def __init__(self, shape, origin=None):
        self.shape = tuple(shape)
        if origin is None:
            origin = (0,) * len(shape)
        self.origin = tuple(origin)

    @staticmethod
    def from_bounds(*bounds):
        """Initialize a box from its bounds

        Parameters
        ----------
        bounds: tuple of (min,max) pairs
            Min/Max coordinate for every dimension

        Returns
        -------
        bbox: :class:`scarlet2.bbox.Box`
            A new box bounded by the input bounds.
        """
        shape = tuple(max(0, cmax - cmin) for cmin, cmax in bounds)
        origin = (cmin for cmin, cmax in bounds)
        return Box(shape, origin=origin)

    @staticmethod
    def from_data(X, min_value=0):
        """Define box where `X` is above `min_value`

        Parameters
        ----------
        X: jnp.ndarray
            Data to threshold
        min_value: float
            Minimum value of the result.

        Returns
        -------
        bbox: :class:`scarlet2.bbox.Box`
            Bounding box for the thresholded `X`
        """
        sel = X > min_value
        if sel.any():
            nonzero = jnp.where(sel)
            bounds = []
            for dim in range(len(X.shape)):
                bounds.append((nonzero[dim].min(), nonzero[dim].max() + 1))
        else:
            bounds = [[0, 0]] * len(X.shape)
        return Box.from_bounds(*bounds)

    def contains(self, p):
        """Whether the box contains a given coordinate `p`
        """
        if len(p) != self.D:
            raise ValueError(f"Dimension mismatch in {p} and {self.D}")

        for d in range(self.D):
            if p[d] < self.origin[d] or p[d] >= self.origin[d] + self.shape[d]:
                return False
        return True

    def insert_into(self, image, sub):
        """Insert `sub` into `image` according to this bbox

        Inverse operation to :func:`~scarlet.bbox.Box.extract_from`.

        Parameters
        ----------
        image: array
            Full image
        sub: array
            Extracted sub-image

        Returns
        -------
        image: array
        """
        imbox = Box(image.shape)

        im_slices, sub_slices = overlap_slices(imbox, self)
        image[im_slices] = sub[sub_slices]
        return image


    @property
    def D(self):
        """Dimensionality of this BBox
        """
        return len(self.shape)

    @property
    def start(self):
        """Tuple of start coordinates
        """
        return self.origin

    @property
    def stop(self):
        """Tuple of stop coordinates
        """
        return tuple(o + s for o, s in zip(self.origin, self.shape))

    @property
    def center(self):
        """Tuple of center coordinates
        """
        return tuple(o + s // 2 for o, s in zip(self.origin, self.shape))

    @property
    def bounds(self):
        """Bounds of the box
        """
        return tuple((o, o + s) for o, s in zip(self.origin, self.shape))

    @property
    def slices(self):
        """Bounds of the box as slices
        """
        return tuple([slice(o, o + s) for o, s in zip(self.origin, self.shape)])

    @property
    def spatial(self):
        """Spatial component of higher-dimensional box"""
        assert self.D >= 2
        return self[-2:]

    def set_center(self, pos):
        """Center box at given position
        """
        pos_ = tuple(_.item() for _ in pos)
        origin = tuple(o + p - c for o, p, c in zip(self.origin, pos_, self.center))
        object.__setattr__(self, 'origin', origin)

    def grow(self, delta):
        """Grow the Box by the given delta in each direction
        """
        if not hasattr(delta, "__iter__"):
            delta = [delta] * self.D
        origin = tuple([self.origin[d] - delta[d] for d in range(self.D)])
        shape = tuple([self.shape[d] + 2 * delta[d] for d in range(self.D)])
        return Box(shape, origin=origin)

    def shrink(self, delta):
        """Shrink the Box by the given delta in each direction
        """
        if not hasattr(delta, "__iter__"):
            delta = [delta] * self.D
        origin = tuple([self.origin[d] + delta[d] for d in range(self.D)])
        shape = tuple([self.shape[d] - 2 * delta[d] for d in range(self.D)])
        return Box(shape, origin=origin)

    def __or__(self, other):
        """Union of two bounding boxes

        Parameters
        ----------
        other: `Box`
            The other bounding box in the union

        Returns
        -------
        result: `Box`
            The smallest rectangular box that contains *both* boxes.
        """
        if other.D != self.D:
            raise ValueError(f"Dimension mismatch in the boxes {other} and {self}")
        bounds = []
        for d in range(self.D):
            bounds.append(
                (min(self.start[d], other.start[d]), max(self.stop[d], other.stop[d]))
            )
        return Box.from_bounds(*bounds)

    def __and__(self, other):
        """Intersection of two bounding boxes

        If there is no intersection between the two bounding
        boxes then an empty bounding box is returned.

        Parameters
        ----------
        other: `Box`
            The other bounding box in the intersection

        Returns
        -------
        result: `Box`
            The rectangular box that is in the overlap region
            of both boxes.
        """
        if other.D != self.D:
            raise ValueError(f"Dimension mismatch in the boxes {other} and {self}")
        assert other.D == self.D
        bounds = []
        for d in range(self.D):
            bounds.append(
                (max(self.start[d], other.start[d]), min(self.stop[d], other.stop[d]))
            )
        return Box.from_bounds(*bounds)

    def __getitem__(self, i):
        s_ = self.shape[i]
        o_ = self.origin[i]
        if not hasattr(s_, "__iter__"):
            s_ = (s_,)
            o_ = (o_,)
        return Box(s_, origin=o_)

    def __add__(self, offset):
        if not hasattr(offset, "__iter__"):
            offset = (offset,) * self.D
        origin = tuple([a + o for a, o in zip(self.origin, offset)])
        return Box(self.shape, origin=origin)

    def __sub__(self, offset):
        if not hasattr(offset, "__iter__"):
            offset = (offset,) * self.D
        origin = tuple([a - o for a, o in zip(self.origin, offset)])
        return Box(self.shape, origin=origin)

    def __matmul__(self, bbox):
        bounds = self.bounds + bbox.bounds
        return Box.from_bounds(*bounds)

    def __copy__(self):
        return Box(self.shape, origin=self.origin)

    def __eq__(self, other):
        return self.shape == other.shape and self.origin == other.origin

    def __hash__(self):
        return hash((self.shape, self.origin))


def overlap_slices(bbox1, bbox2, return_boxes=False):
    """Slices of bbox1 and bbox2 that overlap

    Parameters
    ----------
    bbox1: `~scarlet.bbox.Box`
    bbox2: `~scarlet.bbox.Box`

    Returns
    -------
    slices: tuple of slices
        The slice of an array bounded by `bbox1` and
        the slice of an array bounded by `bbox2` in the
        overlapping region.
    """
    overlap = bbox1 & bbox2
    _bbox1 = overlap - bbox1.origin
    _bbox2 = overlap - bbox2.origin
    if return_boxes:
        return _bbox1, _bbox2
    slices = (
        _bbox1.slices,
        _bbox2.slices,
    )
    return slices



import jax.numpy as jnp
from jax import vmap
import scarlet2
from scarlet2 import *
from numpy.testing import assert_allclose
import astropy.units as u

from scarlet2.measure import get_scale
from scarlet_test_data import data_path
import os
import astropy.io.fits as fits
from astropy.wcs import WCS

T0 = 30
ellipticity = jnp.array((0.3,0.5))
morph = GaussianMorphology(size=T0, ellipticity=ellipticity)

print("Constructed moments")
print(f"Size: {T0}, Ellipticity: {ellipticity}")

g = scarlet2.measure.moments(component=morph(), N=2)

def test_measure_size():
    g = scarlet2.measure.moments(component=morph(), N=2)
    print("Measured moments")
    print(f"Size: {g.size}")
    assert_allclose(T0, g.size, rtol=1e-3)

def test_measure_ellipticity():
    g = scarlet2.measure.moments(component=morph(), N=2)
    print("Measured moments")
    print(f"Ellipticity: {g.ellipticity}")
    assert_allclose(ellipticity, g.ellipticity, rtol=2e-3)

def test_gaussian_from_moments():
    g = scarlet2.measure.moments(component=morph(), N=2)
    # generate Gaussian from moments
    T = g.size
    ellipticity = g.ellipticity
    morph2 = GaussianMorphology(T, ellipticity)
    assert_allclose(morph(), morph2(), rtol=1e-2)

def test_rotate_moments():
    a = 90*u.deg
    a.to(u.deg).value
    # rotate 90 degrees counterclockwise the image
    # and measure its moments
    g2 = scarlet2.measure.moments(jnp.rot90(morph()))

    # rotate moments computed on the original image
    g.rotate(a)
    g.ellipticity

    assert_allclose(g.ellipticity, g2.ellipticity, rtol=1e-6)

def test_resize_moments():
    c = 0.5

    # resize the image
    morph2 = GaussianMorphology(size=T0*c, 
                            ellipticity=ellipticity,
                            shape=morph().shape)
    g2 = scarlet2.measure.moments(morph2(),2)

    # resize moments computed on the original image
    g = scarlet2.measure.moments(component=morph(), N=2)
    g.resize(0.5)

    assert_allclose(g.size, g2.size, rtol=1e-3)

def test_wcs_transfer_moments():

    # Load the HSC image WCS
    obs_hdu = fits.open(os.path.join(data_path, "test_resampling", "Cut_HSC1.fits"))
    wcs_hsc = WCS(obs_hdu[0].header)

    # Load the HST image WCS
    hst_hdu = fits.open(os.path.join(data_path, "test_resampling", "Cut_HST1.fits"))
    wcs_hst = WCS(hst_hdu[0].header)

    # Mock a rotation of 90 deg counter-clockwise of the HST WCS 
    phi = 90 / 180 * jnp.pi # in rad
    R = jnp.array([[jnp.cos(phi), jnp.sin(phi)],
                   [-jnp.sin(phi), jnp.cos(phi)]])

    wcs_hst.wcs.pc = R @ wcs_hst.wcs.pc
    im_hst = jnp.rot90(morph())

    # Generate the same image seen from HSC
    h = (get_scale(wcs_hst) / get_scale(wcs_hsc)).mean()
    T1 = T0 * h 
    ellipticity1 = jnp.array((0.3,0.5))
    morph1 = GaussianMorphology(size=T1, ellipticity=ellipticity1, shape=im_hst.shape)
    im_hsc = morph1()

    # Measure moments of the HST image
    g0 = scarlet2.measure.moments(im_hst)

    # Measure moments of the HSC image
    g1 = scarlet2.measure.moments(im_hsc)

    # Transfer moments from HST to HSC frame
    g0.transfer(wcs_hst, wcs_hsc)

    # Check that size and ellipticity are conserved
    assert_allclose(g1.size, g0.size, rtol=1e-3)
    assert_allclose(g1.ellipticity, g0.ellipticity, rtol=1e-2)

def test_wcs_transfer_w_flip_moments():

    # same as above with RA W->E instead of E->W convention in HST WCS
    # Load the HSC image WCS
    obs_hdu = fits.open(os.path.join(data_path, "test_resampling", "Cut_HSC1.fits"))
    wcs_hsc = WCS(obs_hdu[0].header)

    # Load the HST image WCS
    hst_hdu = fits.open(os.path.join(data_path, "test_resampling", "Cut_HST1.fits"))
    wcs_hst = WCS(hst_hdu[0].header)

    # Mock a rotation of 90 deg counter-clockwise of the HST WCS 
    phi = 90 / 180 * jnp.pi # in rad
    R = jnp.array([[jnp.cos(phi), jnp.sin(phi)],
                   [-jnp.sin(phi), jnp.cos(phi)]])

    wcs_hst.wcs.pc = R @ wcs_hst.wcs.pc
    wcs_hst.wcs.pc *= jnp.array([[-1], [1]])
    
    # rotate and flip x
    im_hst = jnp.rot90(morph())[:,::-1]

    # Generate the same image seen from HSC
    h = (get_scale(wcs_hst) / get_scale(wcs_hsc)).mean()
    T1 = T0 * h 
    ellipticity1 = jnp.array((0.3,0.5))
    morph1 = GaussianMorphology(size=T1, ellipticity=ellipticity1, shape=im_hst.shape)
    im_hsc = morph1()

    # Measure moments of the HST image
    g0 = scarlet2.measure.moments(im_hst)

    # Measure moments of the HSC image
    g1 = scarlet2.measure.moments(im_hsc)

    # Transfer moments from HST to HSC frame
    g0.transfer(wcs_hst, wcs_hsc)

    # Check that size and ellipticity are conserved
    assert_allclose(g1.size, g0.size, rtol=1e-3)
    assert_allclose(g1.ellipticity, g0.ellipticity, rtol=1e-2)

def test_wcs_transfer_moments_multichannels():

    # Load the HSC image WCS
    obs_hdu = fits.open(os.path.join(data_path, "test_resampling", "Cut_HSC1.fits"))
    wcs_hsc = WCS(obs_hdu[0].header)

    # Load the HST image WCS
    hst_hdu = fits.open(os.path.join(data_path, "test_resampling", "Cut_HST1.fits"))
    wcs_hst = WCS(hst_hdu[0].header)

    # Mock a rotation of 90 deg counter-clockwise of the HST WCS 
    phi = 90 / 180 * jnp.pi # in rad
    R = jnp.array([[jnp.cos(phi), jnp.sin(phi)],
                   [-jnp.sin(phi), jnp.cos(phi)]])

    wcs_hst.wcs.pc = R @ wcs_hst.wcs.pc

    nc = 5
    im_hst = jnp.repeat(morph()[None,:,:], repeats=nc, axis=0)
    im_hst = vmap(jnp.rot90)(im_hst)

    # Generate the same image seen from HSC
    h = (get_scale(wcs_hst) / get_scale(wcs_hsc)).mean()
    T1 = T0 * h 
    ellipticity1 = jnp.array((0.3,0.5))
    morph1 = GaussianMorphology(size=T1, ellipticity=ellipticity1, shape=im_hst.shape)
    im_hsc = morph1()

    # Measure moments of the HST image
    g0 = scarlet2.measure.moments(im_hst)

    # Measure moments of the HSC image
    g1 = scarlet2.measure.moments(im_hsc)

    # Transfer moments from HST to HSC frame
    g0.transfer(wcs_hst, wcs_hsc)
    # Check that size and ellipticity are conserved
    assert_allclose(jnp.repeat(g1.size, nc), g0.size, rtol=1e-3)
    assert_allclose(jnp.repeat(g1.ellipticity[:,None], nc, 1), g0.ellipticity, rtol=1e-2)


import jax.numpy as jnp

from .module import Module
from .morphology import GaussianMorphology


class PSF(Module):
    pass


class ArrayPSF(PSF):
    morphology: jnp.ndarray

    def __call__(self):
        return self.morphology / self.morphology.sum((-2,-1), keepdims=True)


class GaussianPSF(PSF):
    morphology: GaussianMorphology

    def __init__(self, sigma):
        self.morphology = GaussianMorphology(sigma)

    def __call__(self):
        morph = self.morphology()
        morph /= morph.sum((-2,-1), keepdims=True)
        return morph



import numpy as jnp
import numpy.ma as ma
import math
import astropy.units as u

from .source import Component
from .frame import get_scale, get_angle, get_sign

def max_pixel(component):
    """Determine pixel with maximum value

    Parameters
    ----------
    component: `scarlet.Component` or array
        Component to analyze or its hyperspectral model
    """
    if isinstance(component, Component):
        model = component()
        origin = jnp.array(component.bbox.origin)
    else:
        model = component
        origin = 0

    return jnp.array(jnp.unravel_index(jnp.argmax(model), model.shape)) + origin


def flux(component):
    """Determine total flux in every channel

    Parameters
    ----------
    component: `scarlet.Component` or array
        Component to analyze or its hyperspectral model
    """
    if isinstance(component, Component):
        model = component()
    else:
        model = component

    return model.sum(axis=(-2, -1))


def centroid(component):
    """Determine spatial centroid of model

    Parameters
    ----------
    component: `scarlet2.Component` or array
        Component to analyze or its hyperspectral model
    """
    if isinstance(component, Component):
        model = component()
        origin = jnp.array(component.bbox.spatial.origin)
    else:
        model = component
        origin = 0, 0

    grid_y, grid_x = jnp.indices(model.shape[-2:])
    if len(model.shape) == 3:
        grid_y = grid_y[None, :, :]
        grid_x = grid_x[None, :, :]
    f = flux(model)
    c = (grid_y * model).sum(axis=(-2, -1)) / f + origin[0], (grid_x * model).sum(axis=(-2, -1)) / f + origin[1]
    return jnp.array(c)


def snr(component, observations):
    """Determine SNR with morphology as weight function

    Parameters
    ----------
    component: `scarlet.Component` or array
        Component to analyze or its hyperspectral model

    observations: `scarlet.Observation` or list thereof
    """
    if not hasattr(observations, "__iter__"):
        observations = (observations,)

    if hasattr(component, "get_model"):
        frame = None
        if not prerender:
            frame = observations[0].model_frame
        model = component.get_model(frame=frame)
    else:
        model = component

    M = []
    W = []
    var = []
    # convolve model for every observation;
    # flatten in channel direction because it may not have all C channels; concatenate
    # do same thing for noise variance
    for obs in observations:
        noise_rms = 1 / jnp.sqrt(ma.masked_equal(obs.weights, 0))
        ma.set_fill_value(noise_rms, jnp.inf)
        model_ = obs.render(model)
        M.append(model_.reshape(-1))
        W.append((model_ / (model_.sum(axis=(-2, -1))[:, None, None])).reshape(-1))
        noise_var = noise_rms ** 2
        var.append(noise_var.reshape(-1))
    M = jnp.concatenate(M)
    W = jnp.concatenate(W)
    var = jnp.concatenate(var)

    # SNR from Erben (2001), eq. 16, extended to multiple bands
    # SNR = (I @ W) / sqrt(W @ Sigma^2 @ W)
    # with W = morph, Sigma^2 = diagonal variance matrix
    snr = (M * W).sum() / jnp.sqrt(((var * W) * W).sum())

    return snr


# moment code adapted from https://github.com/pmelchior/shapelens/blob/master/src/Moments.cc
def moments(component, N=2, center=None, weight=None):
    """Compute centered moments of the component.

    Parameters
    ----------
    component: `scarlet.Component` or array
        Component to analyze or its hyperspectral model
    N: int >=0
        Moment order
    center: array
        2D coordinate in frame of `component`
    weight: array
        weight function with same shape as `component`
    """
    return Moments(component, N=N, center=center, weight=weight)


class Moments(dict):

    def __init__(self, component, N=2, center=None, weight=None):
        """Compute moments of the light distribution.

         Parameters
         ----------
         component: `scarlet.Component` or array
             Component to analyze or its hyperspectral model
         N: int >=0
             Moment order
         center: array
             2D coordinate in frame of `component`
         weight: array
             weight function with same shape as `component`
         """
        super().__init__()

        if isinstance(component, Component):
            model = component()
        else:
            model = component

        if weight is None:
            weight = 1

        grid_y, grid_x = jnp.indices(model.shape[-2:])

        if model.ndim == 3:
            grid_y = grid_y[None, :, :]
            grid_x = grid_x[None, :, :]

        self.N = N
        for n in range(self.N + 1):
            for m in range(n + 1):
                # moments ordered by power in y, then x
                self[m, n - m] = (grid_y ** m * grid_x ** (n - m) * model * weight).sum(axis=(-2, -1))

            if n == 1:
                # shift grid to produce centered moments
                if center is None:
                    center = self[1, 0] / self[0, 0], self[0, 1] / self[0, 0]  # centroid
                    self[1, 0] = jnp.zeros_like(self[1, 0])
                    self[0, 1] = jnp.zeros_like(self[0, 1])
                else:
                    center = jnp.asarray(center)
                    # centroid wrt given center
                    self[1, 0] -= center[0] * self[0, 0]
                    self[0, 1] -= center[1] * self[0, 0]
                if model.ndim == 3 and center[0].ndim == 1:
                    center = center[0][:, None, None], center[1][:, None, None]
                grid_y = grid_y - center[0]
                grid_x = grid_x - center[1]

    @property
    def order(self):
        return max(key[0] for key in self.keys())

    @property
    def centroid(self):
        return jnp.array((self[1, 0] / self[0, 0], self[0, 1] / self[0, 0]))

    @property
    def size(self):
        """Determine size from moments
        """
        flux = self[0, 0]
        T = (self[0, 2] / flux * self[2, 0] / flux - (self[1, 1] / flux) ** 2) ** (1 / 4)
        return T

    @property
    def ellipticity(self):
        """Determine complex ellipticity from moments.

        Returns:
        --------
        jnp.array
        """
        ellipticity = (self[0, 2] - self[2, 0] + 2j * self[1, 1]) / (self[2, 0] + self[0, 2])
        return jnp.array((ellipticity.real, ellipticity.imag))

    def deconvolve(self, p):
        """Deconvolve from moments p"""
        g = self
        Nmin = min(p.order, g.order)

        # use explicit relations for up to 2nd moments
        g[0, 0] /= p[0, 0]
        if Nmin >= 1:
            g[0, 1] -= g[0, 0] * p[0, 1]
            g[1, 0] -= g[0, 0] * p[1, 0]
            g[0, 1] /= p[0, 0]
            g[1, 0] /= p[0, 0]
            if Nmin >= 2:
                g[0, 2] -= g[0, 0] * p[0, 2] + 2 * g[0, 1] * p[0, 1]
                g[1, 1] -= g[0, 0] * p[1, 1] + g[0, 1] * p[1, 0] + g[1, 0] * p[0, 1]
                g[2, 0] -= g[0, 0] * p[2, 0] + 2 * g[1, 0] * p[1, 0]
                if Nmin >= 3:
                    # use general formula
                    for n in range(3, Nmin + 1):
                        for i in range(n + 1):
                            for j in range(n - i):
                                for k in range(i):
                                    for l in range(j):
                                        g[i, j] -= (
                                                binomial(i, k)
                                                * binomial(j, l)
                                                * g[k, l]
                                                * p[i - k, j - l]
                                        )
                                for k in range(i):
                                    g[i, j] -= binomial(i, k) * g[k, j] * p[i - k, 0]
                                for l in range(j):
                                    g[i, j] -= binomial(j, l) * g[i, l] * p[0, j - l]
                        g[i, j] /= p[0, 0]

    def resize(self, c):
        """
        Resize moments given a scaling factor c
        scaling can be different along x and y (c is therefore a list [c1, c2])
        """
        # Teague (1980), eq. 34
        if jnp.isscalar(c):
            for e in self:
                self[e] = self[e] * c**(2+e[0]+e[1])
        elif len(c)==2:
            for e in self:
                self[e] = self[e] * c[0]**(e[0]+1) * c[1]**(e[1]+1)
        else:
            raise AttributeError("c must be a scalar of a list or array of two components")

    def rotate(self, phi):
        """
        Rotate moments given a rotation angle phi (in astropy units)
        """
        # Teague (1980), eq. 36

        assert u.get_physical_type(phi) == "angle" # check that it's an angle with a suitable unit
        phi = phi.to(u.deg).value
        phi = phi * math.pi / 180 # radian

        mu_p = {}
        for n in range(self.N+1):
            for m in range(n+1):
                j = m
                k = n-m
                value = 0
                for r in range(j+1):
                    for s in range(k+1):
                        value += (-1)**(k-s) * binomial(j, r)*binomial(k, s) *\
                              jnp.cos(phi)**(j-r+s) * jnp.sin(phi)**(k+r-s) *\
                                  self[j+k-r-s, r+s]
                mu_p[m, n-m] = value
        
        for e in self:
            self[e] = mu_p[e]
        
    def transfer(self, wcs_in, wcs_out):
        """
        Compute rescaling and rotation from WCSs and apply to moments
        wcs_in: astropy.wcs.Wcsprm
        wcs_out: astropy.wcs.Wcsprm
        """

        if (wcs_in is not None) and (wcs_out is not None):

            # Rescale moments (amplitude rescaling)
            scale_in = get_scale(wcs_in) * 60**2 # arcsec
            scale_out = get_scale(wcs_out) * 60**2 # arcsec
            c = jnp.array(scale_in) / jnp.array(scale_out)
            self.resize(c)

            # Rotate moments
            phi_in = get_angle(wcs_in)
            phi_out = get_angle(wcs_out)
            phi = (phi_out - phi_in)/jnp.pi*180
            self.rotate(phi*u.deg)

            # Flip moments if WCSs don't share the same convention
            sign_in = get_sign(wcs_in)
            sign_out = get_sign(wcs_out)
            self.resize(sign_in*sign_out)


# adapted from  https://github.com/pmelchior/shapelens/blob/src/DEIMOS.cc
def binomial(n, k):
    if k == 0:
        return 1
    if k > n // 2:
        return binomial(n, n - k)
    result = 1
    for i in range(1, k + 1):
        result *= n - i + 1
        result //= i
    return result






class Scenery:
    # static store for context manager
    scene = None


from . import initialization as init
from . import measure
from . import plot
from .bbox import Box
from .frame import Frame
from .module import Parameter, Module, relative_step
from .morphology import Morphology, ArrayMorphology, GaussianMorphology, SersicMorphology, StarletMorphology
from .observation import Observation
from .psf import PSF, ArrayPSF, GaussianPSF
from .scene import Scene
from .source import Component, DustComponent, Source, PointSource
from .spectrum import Spectrum, ArraySpectrum, StaticArraySpectrum, TransientArraySpectrum
from .wavelets import Starlet



import jax
import jax.numpy as jnp
import equinox as eqx

"""
Some of the code to perform interpolation in Fourier space as been adapted from
https://github.com/GalSim-developers/JAX-GalSim/blob/main/jax_galsim/interpolant.py
https://github.com/GalSim-developers/JAX-GalSim/blob/main/jax_galsim/interpolatedimage.py
"""

### Interpolant class

class Interpolant(eqx.Module):
    extent: int
    def __call__(
        self):
        raise NotImplementedError

    def kernel(self, x):
        raise NotImplementedError
    
    def uval(self, u):
        raise NotImplementedError

### Quintic interpolant

class Quintic(Interpolant):
    def __init__(self):
        self.extent = 3

    def f_0_1_q(self, x):
        return 1 + x*x*x * (-95 + 138*x - 55*x*x) / 12

    def f_1_2_q(self, x):
        return (x-1)*(x-2) * (-138 + 348*x - 249*x*x + 55*x*x*x) / 24


    def f_2_3_q(self, x):
        return (x-2)*(x-3)*(x-3)*(-54 + 50*x - 11*x*x) / 24

    def f_3_q(self, x):
        return jnp.zeros_like(x, dtype=x.dtype)

    def kernel(self, x):
        """
        Quintic kernel values in direct space
        """
        x = jnp.abs(x) # quitic kernel is even

        b1 = x <= 1
        b2 = x <= 2
        b3 = x <= 3 

        return jnp.piecewise(
            x, 
            [b1, (~b1) & b2, (~b2) & b3], 
            [self.f_0_1_q, self.f_1_2_q, self.f_2_3_q, self.f_3_q]
            )
    
    def uval(self, u):
        """
        Quintic kernel values in Fourier space
        """
        u = jnp.abs(u)
        s = jnp.sinc(u)
        piu = jnp.pi*u
        c = jnp.cos(piu)
        ssq = s*s
        piusq = piu*piu
        
        return s * ssq * ssq * (s * (55. - 19. * piusq) + 2. * c * (piusq - 27.))

### Lanczos interpolant

class Lanczos(Interpolant):
    
    def __init__(self, n):
        """
        Lanczos interpolant

        Parameters
        ----------
        n: Lanczos order
        """

        self.extent = n
    
    def f_1(x, n):
        """
        Approximation from galsim
        // res = n/(pi x)^2 * sin(pi x) * sin(pi x / n)
        //     ~= (1 - 1/6 pix^2) * (1 - 1/6 pix^2 / n^2)
        //     = 1 - 1/6 pix^2 ( 1 + 1/n^2 )
        """
        px = jnp.pi * x
        temp = 1./6. * px * px
        res = 1. - temp * (1. + 1. / (n * n))
        return res

    def f_2(x, n):
        px = jnp.pi * x
        return n * jnp.sin(px) * jnp.sin(px / n) / px / px

    def lanczos_n(x, n=3):
        """
        Lanczos interpolation kernel in direct space
        """
        small_x = jnp.abs(x) <= 1e-4
        window_n = jnp.abs(x) <= n

        return jnp.piecewise(
            x, 
            [small_x, (~small_x) & window_n], 
            [f_1, f_2, lambda x, n: jnp.array(0)], n
        )

    def kernel(self, x):
        return self.lanczos_n(x, self.extent)

### Resampling function

def resample2d(signal, coords, warp, interpolant=Quintic()):
    """
    Resample a 2-dimensional image using a Lanczos kernel

    Parameters
    ----------
    signal: array
        2d array containing the signal. We assume here that the coordinates of
        the signal
        shape: [Nx, Ny]
    coords: array
        Coordinates on which the signal is sampled.
        shape: [Nx, Ny, 2]
            - x-coordinates are coords[0,:,0]
            - y-coordinates are coords[:,0,1]
    warp: array
        Coordinates on which to resample the signal.
        shape:[nx, ny, 2]
        [ [[0,  0], [0,  1], ...,  [0,  N-1]],
                           [ ... ],
          [[N-1,0], [N-1,1], ...,  [N-1,N  ]] ]
    n: Lanczos order

    Returns
    -------
    resampled_signal: array
    """

    x = warp[..., 0].flatten()
    y = warp[..., 1].flatten()

    coords_x = coords[0, :, 0]
    coords_y = coords[:, 0, 1]

    h = coords_x[1] - coords_x[0]

    xi = jnp.floor((x - coords_x[0]) / h).astype(jnp.int32)
    yi = jnp.floor((y - coords_y[0]) / h).astype(jnp.int32)

    Nx = coords.shape[0]
    Ny = coords.shape[1]

    def body_fun_x(i, args):
        res, yind, ky, masky = args

        xind = xi + i
        maskx = (xind >= 0) & (xind < Ny)

        kx = interpolant.kernel((x - coords_x[xind]) / h)

        k = kx * ky
        mask = maskx & masky
        res += jnp.where(mask, k * signal[yind, xind], 0)

        return res, yind, ky, masky

    def body_fun_y(i, args):
        res = args

        yind = yi + i
        masky = (yind >= 0) & (yind < Nx)

        ky = interpolant.kernel((y - coords_y[yind]) / h)

        res = jax.lax.fori_loop(-interpolant.extent, interpolant.extent + 1, body_fun_x, (res, yind, ky, masky))[0]

        return res

    res = jax.lax.fori_loop(
        -interpolant.extent, interpolant.extent + 1, body_fun_y, jnp.zeros_like(x).astype(signal.dtype)
    )

    return res.reshape(warp[..., 0].shape)

def resample_hermitian(signal, warp, x_min, y_min, interpolant=Quintic()):
    """
    Resample a 2-dimensional image using a Lanczos kernel
    This is assuming that the signal is Hermitian and starting at 0 on axis=2,
    i.e. f(-x, -y) = conjugate(f(x, y))

    Parameters
    ----------
    signal: array
        2d array containing the signal. We assume here that the coordinates of
        the signal
        shape: [Nx, Ny]
    warp: array
        Coordinates on which to resample the signal, in the grid of signal
        coordinates [[0 ... signal.shape[0]], [0 ... signal.shape[1]]
        shape:[nx, ny, 2]
        [ [[0,  0], [0,  1], ...,  [0,  N-1]],
                           [ ... ],
          [[N-1,0], [N-1,1], ...,  [N-1,N  ]] ]
    n: Lanczos order

    Returns
    -------
    resampled_signal: array
    """

    x = warp[..., 0].flatten()
    y = warp[..., 1].flatten()

    xi = jnp.floor(x-x_min).astype(jnp.int32)
    yi = jnp.floor(y-y_min).astype(jnp.int32)

    xp = xi + x_min
    yp = yi + y_min

    nkx_2 = signal.shape[1]-1
    nkx = signal.shape[0]

    def body_fun_x(i, args):
        res, yind, ky = args

        xind = (xi + i) % nkx
        
        kx = interpolant.kernel(x - (xp+i))

        k = kx * ky
        
        tmp = jnp.where(xind < nkx_2,
                        signal[(nkx - yind) % nkx, nkx - xind - nkx_2].conjugate(),
                        signal[yind, xind - nkx_2]
                        )
        
        res += tmp * k

        return res, yind, ky

    def body_fun_y(i, args):
        res = args

        yind = yi + i

        ky = interpolant.kernel(y - (yp+i))

        res = jax.lax.fori_loop(-interpolant.extent, interpolant.extent + 1, body_fun_x, (res, yind, ky))[0]

        return res

    res = jax.lax.fori_loop(
        -interpolant.extent, interpolant.extent + 1, body_fun_y, jnp.zeros_like(x).astype(signal.dtype)
    )

    return res.reshape(warp[..., 0].shape)

def resample_ops(kimage, shape_in, shape_out, res_in, res_out, phi=None, flip_sign=None, interpolant=Quintic()):
    """
    Resampling operation used in the Multiresolution Renderers
    This is assuming that the signal is Hermitian and starting at 0 on axis=2,
    i.e. f(-x, -y) = conjugate(f(x, y))
    """
    
    # Apply rescaling to the frequencies
    # [0, Fe/2+1]
    # [-Fe/2+1, Fe/2]
    kcoords_out = jnp.stack(jnp.meshgrid(
        jnp.linspace(0, 
                     shape_in / 2 / res_out * res_in, 
                     shape_out // 2 + 1),
        jnp.linspace(-shape_in / 2 / res_out * res_in, 
                     shape_in / 2 / res_out * res_in, 
                     shape_out)
        ), -1)

    
    # Apply rotation to the frequencies
    if phi is not None:
        R = jnp.array([[jnp.cos(phi), jnp.sin(phi)],
                       [-jnp.sin(phi), jnp.cos(phi)]])
        
        b_shape = kcoords_out.shape
        kcoords_out = (R @ kcoords_out.reshape((-1, 2)).T).T.reshape((b_shape))

    k_resampled = jax.vmap(resample_hermitian, in_axes=(0,None,None,None,None))(
        kimage,
        kcoords_out,
        -shape_in/2,
        -shape_in/2,
        interpolant
        )
    
    kx = jnp.linspace(0, jnp.pi, shape_out//2 + 1) * res_in/res_out
    ky = jnp.linspace(-jnp.pi, jnp.pi, shape_out)
    coords = jnp.stack(jnp.meshgrid(kx, ky),-1) / 2 / jnp.pi

    xint_val = interpolant.uval(coords[...,0]) * interpolant.uval(coords[...,1])
    
    return k_resampled * jnp.expand_dims(xint_val, 0)


import os
import warnings
warnings.filterwarnings('ignore')

from numpy.testing import assert_allclose

import scarlet2
from utils import import_scarlet_test_data
import_scarlet_test_data()
from scarlet_test_data import data_path, tests_path

import astropy.io.fits as fits
from astropy.wcs import WCS

import jax
import jax.numpy as jnp

# Load the HSC image data
obs_hdu = fits.open(os.path.join(data_path, "test_resampling", "Cut_HSC1.fits"))
                    
data_hsc = jnp.array(obs_hdu[0].data.byteswap().newbyteorder(), jnp.float32)
wcs_hsc = WCS(obs_hdu[0].header)
channels_hsc = ['g','r','i','z','y']

# Load the HSC PSF data
psf_hsc_data  = fits.open(os.path.join(data_path, "test_resampling", "PSF_HSC.fits"))[0].data.astype('float32')
                    
Np1, Np2 = psf_hsc_data[0].shape
psf_hsc_data = jnp.pad(psf_hsc_data, ((0,0), (1,0), (1,0)))
psf_hsc = scarlet2.ArrayPSF(psf_hsc_data)

# Load the HST image data
hst_hdu = fits.open(os.path.join(data_path, "test_resampling", "Cut_HST1.fits"))

data_hst = hst_hdu[0].data
wcs_hst = WCS(hst_hdu[0].header)
channels_hst = ['F814W']

# Load the HST PSF data
psf_hst = fits.open(os.path.join(data_path, "test_resampling", "PSF_HST.fits"))[0].data
psf_hst = jnp.array(psf_hst[None, :, :], jnp.float32)
psf_hst = jnp.pad(psf_hst, ((0, 0), (1, 0), (1, 0)))
psf_hst_ = jnp.repeat(psf_hst, 5, 0)

psf_hst = scarlet2.ArrayPSF(psf_hst_)

# Scale the HST data
n1, n2 = jnp.shape(data_hst)
data_hst = data_hst.reshape(1, n1, n2).byteswap().newbyteorder()
data_hst *= data_hsc.max() / data_hst.max()

r, N1, N2 = data_hsc.shape

# define two observation packages and match to frame
obs_hst = scarlet2.Observation(data_hst[:1, ...],
                               wcs=wcs_hst,
                               psf=psf_hst,
                               channels=['channel'],
                               weights=None)

obs_hsc = scarlet2.Observation(data_hsc[:1,...],
                              wcs=wcs_hsc,
                              psf=psf_hsc,
                              channels=['channel'],
                              weights=None)

# Building a Frame from hst obs

hst_frame = scarlet2.Frame(
                bbox=scarlet2.Box(shape=data_hst.shape),
                channels=['channel'],
                psf=psf_hst,
                wcs=wcs_hst
)

def test_hst_to_hsc_against_galsim():

    # Initializing renderers
    obs_hsc.match(hst_frame)

    # Deconvolution, Resampling and Reconvolution
    hst_resampled = obs_hsc.render(data_hst)

    # Perform the same operations with galsim 
    out_galsim = jnp.load(os.path.join(tests_path, "galsim_hst_to_hsc_resolution.npy"))

    assert_allclose(out_galsim, hst_resampled[0], atol=1.3e-4)

def test_hst_to_hsc_against_galsim_rotated_wcs():
    # Remember images coordinates are [y, x]
    # Update CRPIX for the 90-degree clockwise rotation
    crpix1_new = n1 - wcs_hst.wcs.crpix[1]
    crpix2_new = wcs_hst.wcs.crpix[0]
    wcs_hst.wcs.crpix = [crpix1_new, crpix2_new]

    # # Mock a rotation of 90 deg clockwise of the HST WCS 
    phi = -90 / 180 * jnp.pi # in rad
    R = jnp.array([[jnp.cos(phi), jnp.sin(phi)],
                    [-jnp.sin(phi), jnp.cos(phi)]])

    data_hst_ = jax.vmap(jnp.rot90)(data_hst)
    psf_hst = jax.vmap(jnp.rot90)(psf_hst_)
    psf_hst = scarlet2.ArrayPSF(psf_hst)

    wcs_hst.wcs.pc = R @ wcs_hst.wcs.pc    

    hst_frame = scarlet2.Frame(
                bbox=scarlet2.Box(shape=data_hst_.shape),
                channels=['channel'],
                psf=psf_hst,
                wcs=wcs_hst
    )

    # Automatically find the difference between observation and model WCSs
    obs_hsc.match(hst_frame)
    
    # Deconvolution, Resampling and Reconvolution
    hst_resampled = obs_hsc.render(data_hst_)

    # Perform the same operations with galsim 
    out_galsim = jnp.load(os.path.join(tests_path, "galsim_hst_to_hsc_resolution.npy"))

    assert_allclose(out_galsim, hst_resampled[0], atol=1.3e-4)

if __name__=="__main__":
    test_hst_to_hsc_against_galsim()
    test_hst_to_hsc_against_galsim_rotated_wcs()


import operator
import math

import jax.lax
import jax.numpy as jnp
from scipy import fftpack


def transform(image, fft_shape, axes=None):
    """The FFT of an image for a given `fft_shape` along desired `axes`

    Parameters
    ----------
    image: array
        The real-space image.
    fft_shape: tuple
        "Fast" shape of the image used to generate the FFT.
    axes: int or tuple
        The dimension(s) of the array that will be transformed.
    """

    # TODO: allow fft_shape = None to automatically determine it

    if axes is None:
        axes = range(len(image.shape))
    else:
        try:
            iter(axes)
        except TypeError:
            axes = (axes,)

    if len(fft_shape) != len(axes):
        msg = (
            "fft_shape self.axes must have the same number of dimensions, got {0}, {1}"
        )
        raise ValueError(msg.format(fft_shape, axes))

    image = _pad(image, fft_shape, axes)
    image = jnp.fft.ifftshift(image, axes)
    image_fft = jnp.fft.rfftn(image, axes=axes)
    return image_fft


def inverse(image_fft, fft_shape, image_shape, axes=None):
    """Generate image from its FFT

    Parameters
    ----------
    image_fft: array
        The FFT of the image.
    fft_shape: tuple
        "Fast" shape of the image used to generate the FFT.
    image_shape: tuple
        The actual shape of the image *before padding* and using fast shapes.
        This will regenerate the image with the extra padding trimmed off.
    axes: int or tuple
        The dimension(s) of the array that will be transformed.
    """

    # TODO: allow fft_shape = None to automatically determine it

    if axes is None:
        axes = range(len(image_fft.shape))
    else:
        try:
            iter(axes)
        except TypeError:
            axes = (axes,)

    image = jnp.fft.irfftn(image_fft, fft_shape, axes=axes)
    # Shift the center of the image from the bottom left to the center
    image = jnp.fft.fftshift(image, axes=axes)
    # Trim the image to remove the padding added
    # to reduce fft artifacts
    image = _trim(image, image_shape)
    return image


def convolve(image, kernel, padding=3, axes=None, fft_shape=None, return_fft=False):
    """Convolve image with a kernel

    Parameters
    ----------
    image: array
        Image array
    kernel: array
        Convolution kernel array
    padding: int
        Additional padding to use when generating the FFT
        to supress artifacts.
    axes: tuple or None
        Axes that contain the spatial information for the PSFs.
    """
    return _kspace_op(
        image,
        kernel,
        operator.mul,
        padding=padding,
        axes=axes,
        fft_shape=fft_shape,
        return_fft=return_fft,
    )


def deconvolve(image, kernel, padding=3, axes=None, fft_shape=None, return_fft=False):
    """Deconvolve image with a kernel

    This is usually unstable. Treat with caution!

    Parameters
    ----------
    image: array
        Image array
    kernel: array
        Convolution kernel array
    padding: int
        Additional padding to use when generating the FFT
        to supress artifacts.
    axes: tuple or None
        Axes that contain the spatial information for the PSFs.
    """

    return _kspace_op(
        image,
        kernel,
        operator.truediv,
        padding=padding,
        fft_shape=fft_shape,
        axes=axes,
        return_fft=return_fft,
    )


def _kspace_op(
    image, kernel, f, padding=3, axes=None, fft_shape=None, return_fft=False
):
    if axes is None:
        axes = range(len(image.shape))
    else:
        try:
            iter(axes)
        except TypeError:
            axes = (axes,)

    # assumes kernel FFT has been computed with large enough shape to cover also image
    if kernel.dtype in (jnp.complex64, jnp.complex128):
        fft_shape = [kernel.shape[ax] for ax in axes]
        fft_shape[-1] = 2 * (
            fft_shape[-1] - 1
        )  # real-valued FFT has 1/2 of the frequencies
        kernel_fft = kernel
    else:
        if fft_shape is None:
            fft_shape = _get_fast_shape(
                image.shape, kernel.shape, padding=padding, axes=axes
            )
        kernel_fft = transform(kernel, fft_shape, axes=axes)

    image_fft = transform(image, fft_shape, axes=axes)
    image_fft_ = f(image_fft, kernel_fft)
    if return_fft:
        return image_fft_
    image_ = inverse(image_fft_, fft_shape, image.shape, axes=axes)
    return image_


def _get_fast_shape(im_or_shape1, im_or_shape2, axes=None, padding=3, max_shape=False):
    """Return the fast fft shapes for each spatial axis

    Calculate the fast fft shape for each dimension in
    axes.
    """
    if hasattr(im_or_shape1, "shape"):
        shape1 = im_or_shape1.shape
    else:
        shape1 = im_or_shape1
    if hasattr(im_or_shape2, "shape"):
        shape2 = im_or_shape2.shape
    else:
        shape2 = im_or_shape2

    # Make sure the shapes are the same size
    if len(shape1) != len(shape2):
        msg = (
            "img1 and img2 must have the same number of dimensions, but got {0} and {1}"
        )
        raise ValueError(msg.format(len(shape1), len(shape2)))

    if axes is None:
        axes = range(len(shape1))
    else:
        try:
            iter(axes)
        except TypeError:
            axes = (axes,)

    # Set the combined shape based on the total dimensions
    combine_shapes = lambda s1, s2: max(s1, s2) if max_shape else s1 + s2
    shape = [combine_shapes(shape1[ax], shape2[ax]) + padding for ax in axes]
    # Use the next fastest shape in each dimension
    # TODO: check what jnp actually uses for FFTs
    shape = [fftpack.helper.next_fast_len(s) for s in shape]
    return shape


def _trim(arr, newshape):
    """Return the center newshape portion of the array.

    Note: If the array shape is odd and the target is even,
    the center of `arr` is shifted to the center-right
    pixel position.
    This is slightly different than the scipy implementation,
    which uses the center-left pixel for the array center.
    The reason for the difference is that we have
    adopted the convention of `np.fft.fftshift` in order
    to make sure that changing back and forth from
    fft standard order (0 frequency and position is
    in the bottom left) to 0 position in the center.
    """
    startind = tuple((c - s + 1) // 2 for c, s in zip(arr.shape, newshape))
    return jax.lax.dynamic_slice(arr, startind, newshape)


def _fast_zero_pad(arr, pad_width):
    """Fast version of numpy.pad when `mode="constant"`

    Executing `numpy.pad` with zeros is ~1000 times slower
    because it doesn't make use of the `zeros` method for padding.

    Parameters
    ---------
    arr: array
        The array to pad
    pad_width: tuple
        Number of values padded to the edges of each axis.
        See numpy docs for more.

    Returns
    -------
    result: array
        The array padded with `constant_values`
    """
    newshape = tuple([a + ps[0] + ps[1] for a, ps in zip(arr.shape, pad_width)])
    result = jnp.zeros(newshape, dtype=arr.dtype)
    start = tuple(start for start, end in pad_width)
    result = jax.lax.dynamic_update_slice(result, arr, start_indices=start)
    return result


def _pad(arr, newshape, axes=None, mode="constant", constant_values=0):
    """Pad an array to fit into newshape

    Pad `arr` with zeros to fit into newshape,
    which uses the `np.fft.fftshift` convention of moving
    the center pixel of `arr` (if `arr.shape` is odd) to
    the center-right pixel in an even shaped `newshape`.
    """
    if axes is None:
        newshape = jnp.asarray(newshape)
        currshape = jnp.array(arr.shape)
        dS = newshape - currshape
        startind = (dS + 1) // 2
        endind = dS - startind
        pad_width = list(zip(startind, endind))
    else:
        # only pad the axes that will be transformed
        pad_width = [(0, 0) for axis in arr.shape]
        try:
            len(axes)
        except TypeError:
            axes = [axes]
        for a, axis in enumerate(axes):
            dS = newshape[a] - arr.shape[axis]
            startind = (dS + 1) // 2
            endind = dS - startind
            pad_width[axis] = (startind, endind)

    # if mode == "constant" and constant_values == 0:   
    # result = _fast_zero_pad(arr, pad_width)
    # else:
    result = jnp.pad(arr, pad_width, mode=mode)
    return result


def good_fft_size(input_size):
    # Code from JAX-Galsim
    # https://github.com/GalSim-developers/JAX-GalSim/blob/4b12d6b3520938cd823ae3978c400bb9a2b454d3/jax_galsim/image.py#L830
    # Reference from GalSim C++
    # https://github.com/GalSim-developers/GalSim/blob/ece3bd32c1ae6ed771f2b489c5ab1b25729e0ea4/src/Image.cpp#L1009
    # Reduce slightly to eliminate potential rounding errors:
    insize = (1.0 - 1.0e-5) * input_size
    log2n = math.log(2.0) * math.ceil(math.log(insize) / math.log(2.0))
    log2n3 = math.log(3.0) + math.log(2.0) * math.ceil(
        (math.log(insize) - math.log(3.0)) / math.log(2.0)
    )
    log2n3 = max(log2n3, math.log(6.0))  # must be even number
    Nk = max(int(math.ceil(math.exp(min(log2n, log2n3)) - 1.0e-5)), 2)
    return Nk

def wrap_hermitian_x(im, im_xmin, im_ymin, wrap_xmin, wrap_ymin, wrap_nx, wrap_ny):
    """
    Bernstein & Gruen (2014) arxiv:1401.2636
    This function is taken from JAX-Galsim wrap_image utils written by Matthew R. Becker
    https://github.com/GalSim-developers/JAX-GalSim/blob/4b12d6b3520938cd823ae3978c400bb9a2b454d3/jax_galsim/core/wrap_image.py#L6C1-L54C40
    """

    def wrap_nonhermitian(im, xmin, ymin, nxwrap, nywrap):
        def _body_j(j, vals):
            i, im = vals

            ii = (i - ymin) % nywrap + ymin
            jj = (j - xmin) % nxwrap + xmin

            im = jax.lax.cond(
                # weird way to say if ii != i and jj != j
                # I tried other ways and got test failures
                jnp.abs(ii - i) + jnp.abs(jj - j) != 0,
                lambda im, i, j, ii, jj: im.at[ii, jj].add(im[i, j]),
                lambda im, i, j, ii, jj: im,
                im,
                i,
                j,
                ii,
                jj,
            )

            return [i, im]

        def _body_i(i, vals):
            im = vals
            _, im = jax.lax.fori_loop(0, im.shape[1], _body_j, [i, im])
            return im

        im = jax.lax.fori_loop(0, im.shape[0], _body_i, im)
        return im


    def expand_hermitian_x(im):
        return jnp.concatenate([im[:, 1:][::-1, ::-1].conjugate(), im], axis=1)


    def contract_hermitian_x(im):
        return im[:, im.shape[1] // 2 :]

    
    im_exp = expand_hermitian_x(im)
    im_exp = wrap_nonhermitian(
        im_exp, wrap_xmin - im_xmin, wrap_ymin - im_ymin, wrap_nx, wrap_ny
    )
    return contract_hermitian_x(im_exp)


import operator

import equinox as eqx
import jax
import jax.numpy as jnp
from astropy.coordinates import SkyCoord

from . import Scenery
from .bbox import Box, overlap_slices
from .module import Module
from .morphology import Morphology
from .spectrum import Spectrum


class Component(Module):
    center: jnp.ndarray
    spectrum: Spectrum
    morphology: Morphology
    bbox: Box = eqx.field(static=True, init=False)

    def __init__(self, center, spectrum, morphology):
        assert isinstance(spectrum, Spectrum)
        self.spectrum = spectrum
        assert isinstance(morphology, Morphology)
        self.morphology = morphology

        if isinstance(center, SkyCoord):
            try:
                center = Scenery.scene.frame.get_pixel(center)
            except AttributeError:
                print("`center` defined in sky coordinates can only be created within the context of a Scene")
                print("Use 'with Scene(frame) as scene: (...)'")
                raise
        self.center = center

        box = Box(spectrum.shape)
        box2d = Box(morphology.shape)
        box2d.set_center(center.astype(int))
        self.bbox = box @ box2d

    def __call__(self):
        # Boxed and centered model
        delta_center = (self.bbox.center[-2] - self.center[-2], self.bbox.center[-1] - self.center[-1])
        return self.spectrum()[:, None, None] * self.morphology(delta_center=delta_center)[None, :, :]


class DustComponent(Component):
    def __call__(self):
        return jnp.exp(-super().__call__())


class Source(Component):
    components: list
    component_ops: list = eqx.field(static=True)

    def __init__(self, center, spectrum, morphology):
        # set the base component
        super().__init__(center, spectrum, morphology)
        # create the empty component list
        self.components = list()
        self.component_ops = list()

        # add this source to the active scene
        try:
            Scenery.scene.sources.append(self)
        except AttributeError:
            print("Source can only be created within the context of a Scene")
            print("Use 'with Scene(frame) as scene: Source(...)'")
            raise

    def add_component(self, component, op):
        assert isinstance(component, (Source, Component))

        # if component is a source, it's already registered in scene
        # remove it from scene to not call it twice
        if isinstance(component, Source):
            try:
                Scenery.scene.sources.remove(component)
            except AttributeError:
                print("Source can only be modified within the context of a Scene")
                print("Use 'with Scene(frame) as scene: Source(...)'")
                raise
            except ValueError:
                pass

        # adding a full source will maintain its ownership of components:
        # hierarchical definition of sources withing sources
        self.components.append(component)
        self.component_ops.append(op)
        return self

    def __iadd__(self, component):
        return self.add_component(component, operator.add)

    def __imul__(self, component):
        return self.add_component(component, operator.mul)

    def __call__(self):
        base = super()
        model = base.__call__()
        for component, op in zip(self.components, self.component_ops):
            model_ = component()
            # cut out regions from model and model_
            bbox, bbox_ = overlap_slices(base.bbox, component.bbox, return_boxes=True)
            sub_model = jax.lax.dynamic_slice(model, bbox.start, bbox.shape)
            sub_model_ = jax.lax.dynamic_slice(model_, bbox_.start, bbox_.shape)
            # combine with operator
            sub_model = op(sub_model, sub_model_)
            # add model_ back in full model
            model = jax.lax.dynamic_update_slice(model, sub_model, bbox.start)
        return model


class PointSource(Source):
    def __init__(self, center, spectrum):
        try:
            frame = Scenery.scene.frame
        except AttributeError:
            print("Source can only be created within the context of a Scene")
            print("Use 'with Scene(frame) as scene: Source(...)'")
            raise
        if frame.psf is None:
            raise AttributeError("PointSource can only be create with a PSF in the model frame")
        morphology = frame.psf.morphology

        super().__init__(center, spectrum, morphology)



import equinox as eqx
import jax.numpy as jnp
import jax

from .fft import convolve, deconvolve, _get_fast_shape, transform, good_fft_size, _trim, _pad

from .interpolation import resample_ops

from .fft import wrap_hermitian_x

from .measure import get_angle, get_sign

class Renderer(eqx.Module):
    def __call__(
        self, model, key=None
    ):  # key is needed to chain renderers with eqx.nn.Sequential
        raise NotImplementedError


class NoRenderer(Renderer):
    def __call__(self, model, key=None):
        return model


class ChannelRenderer(Renderer):
    channel_map: (None, list, slice) = None

    def __init__(self, model_frame, obs_frame):
        if obs_frame.channels == model_frame.channels:
            channel_map = None
        else:
            try:
                channel_map = [
                    list(model_frame.channels).index(c)
                    for c in list(obs_frame.channels)
                ]
            except ValueError:
                msg = "Cannot match channels between model and observation.\n"
                msg += f"Got {model_frame.channels} and {obs_frame.channels}."
                raise ValueError(msg)

            min_channel = min(channel_map)
            max_channel = max(channel_map)
            if max_channel + 1 - min_channel == len(channel_map):
                channel_map = slice(min_channel, max_channel + 1)
        self.channel_map = channel_map

    def __call__(self, model, key=None):
        """Map model channels onto the observation channels

        Parameters
        ----------
        model: array
            The hyperspectral model
        Returns
        -------
        obs_model: array
            `model` mapped onto the observation channels
        """
        if self.channel_map is None:
            return model
        if isinstance(self.channel_map, (slice, list)):
            return model[self.channel_map, :, :]
        # not yet used by any renderer: full matrix mapping between model and observation channels
        return jnp.dot(self.channel_map, model)


class ConvolutionRenderer(Renderer):
    def __init__(self, model_frame, obs_frame):
        # create PSF model
        psf = model_frame.psf()
        if len(psf.shape) == 2:  # only one image for all bands
            psf_model = jnp.tile(psf, (obs_frame.bbox.shape[0], 1, 1))
        else:
            psf_model = psf
        
        # make sure fft uses a shape large enough to cover the convolved model
        fft_shape = _get_fast_shape(
            model_frame.bbox.shape, psf_model.shape, padding=3, axes=(-2, -1)
        )

        # compute and store diff kernel in Fourier space
        diff_kernel_fft = deconvolve(
            obs_frame.psf(),
            psf_model,
            axes=(-2, -1),
            fft_shape=fft_shape,
            return_fft=True,
        )
        object.__setattr__(self, "_diff_kernel_fft", diff_kernel_fft)

    def __call__(self, model, key=None):
        return convolve(model, self._diff_kernel_fft, axes=(-2, -1))

"""
Preprocess:
    - padd img, psf_in and psf_out on the according goodfftsize
    - return kimages

Resample:
    - resample the three kimages on the target kgrid
    - return these kimages

Postprocess:
    - Deconvolve model PSF and convolve obs PSF in Fourier space
    - kwrapping
    - ifft and cropping to obs frame
"""

class PreprocessMultiresRenderer(Renderer):
    """
    Perform padding and Fourier transform of the model image, model psf and
    observed psf
    """

    def __init__(self, model_frame, obs_frame, padding=None):

        if padding == None:
            # use 4 times padding
            padding = 4
        object.__setattr__(self, "_padding", padding)

        # create PSF model
        psf_model = model_frame.psf()
        
        if len(psf_model.shape) == 2:  # only one image for all bands
            psf_model = jnp.tile(psf_model, (obs_frame.bbox.shape[0], 1, 1))
        
        object.__setattr__(self, "_psf_model", psf_model)

        psf_obs = obs_frame.psf()

        object.__setattr__(self, "_psf_obs", psf_obs)

        fft_shape_model_im = good_fft_size(self._padding*max(model_frame.bbox.shape))
        fft_shape_model_psf = good_fft_size(self._padding*max(psf_model.shape))
        fft_shape_obs_psf = good_fft_size(self._padding*max(psf_obs.shape))

        object.__setattr__(self, "fft_shape_model_im", fft_shape_model_im)
        object.__setattr__(self, "fft_shape_model_psf", fft_shape_model_psf)
        object.__setattr__(self, "fft_shape_obs_psf", fft_shape_obs_psf)

    def __call__(self, model, key=None):

        psf_model = self._psf_model
        psf_obs = self._psf_obs

        model_kim = jnp.fft.fftshift(
            transform(model, (self.fft_shape_model_im, self.fft_shape_model_im), (-2, -1)),
            (-2))
        model_kpsf = jnp.fft.fftshift(
            transform(psf_model, (self.fft_shape_model_psf, self.fft_shape_model_psf), (-2, -1)),
            (-2))
        obs_kpsf = jnp.fft.fftshift(
            transform(psf_obs, (self.fft_shape_obs_psf, self.fft_shape_obs_psf), (-2, -1)),
            (-2))

        return model_kim, model_kpsf, obs_kpsf
    
class ResamplingMultiresRenderer(Renderer):
    """
    Perform the interpolation of the model, model psf and obs psf on the same
    target grid
    """

    def __init__(self, model_frame, obs_frame, padding=None):
        
        if padding == None:
            # use 4 times padding
            padding = 4
        object.__setattr__(self, "_padding", padding)

        fft_shape_model_im = good_fft_size(self._padding*max(model_frame.bbox.shape))
        fft_shape_obs_psf = good_fft_size(self._padding*max(obs_frame.psf().shape))

        # getting the smallest grid to perform the interpolation
        # odd shape is required for k-wrapping later
        fft_shape_target = min(fft_shape_model_im, fft_shape_obs_psf) + 1
        object.__setattr__(self, "fft_shape_target", fft_shape_target)

        object.__setattr__(self, "res_in", model_frame.pixel_size)
        object.__setattr__(self, "res_out", obs_frame.pixel_size)

        # Extract rotation angle between WCSs using jacobian matrices
        angle_in = get_angle(model_frame.wcs)
        angle_out = get_angle(obs_frame.wcs)
        if angle_out - angle_in == 0:
            object.__setattr__(self, "rotation_angle", None)
        else:
            object.__setattr__(self, "rotation_angle", angle_out - angle_in)
        
        # Get flip sign between WCSs using jacobian matrices
        sign_in = get_sign(model_frame.wcs)
        sign_out = get_sign(obs_frame.wcs)
        if (sign_in != sign_out).any():
            raise ValueError("model and observation WCSs have different sign conventions, which is not yet handled by scarlet2")

        object.__setattr__(self, "flip_sign", sign_in*sign_out)

    def __call__(self, kimages, key=None):

        model_kim, model_kpsf, obs_kpsf = kimages

        model_kim_interp = resample_ops(model_kim, model_kim.shape[-2], 
                                        self.fft_shape_target, self.res_in, self.res_out,
                                        phi=self.rotation_angle,
                                        flip_sign=self.flip_sign)

        model_kpsf_interp = resample_ops(model_kpsf, model_kpsf.shape[-2], 
                                        self.fft_shape_target, self.res_in, self.res_out,
                                        phi=self.rotation_angle,
                                        flip_sign=self.flip_sign)
        
        obs_kpsf_interp = resample_ops(obs_kpsf, obs_kpsf.shape[-2], 
                                        self.fft_shape_target, self.res_out, self.res_out)
        
        return model_kim_interp, model_kpsf_interp, obs_kpsf_interp
    

class PostprocessMultiresRenderer(Renderer):
    def __init__(self, model_frame, obs_frame, padding=None):
        if padding == None:
            # use 4 times padding
            padding = 4
        object.__setattr__(self, "_padding", padding)

        fft_shape_model_im = good_fft_size(self._padding*max(model_frame.bbox.shape))
        fft_shape_obs_psf = good_fft_size(self._padding*max(obs_frame.psf().shape))
        object.__setattr__(self, "real_shape_target", obs_frame.bbox.shape)
        
        # getting the smallest grid to perform the interpolation
        # odd shape is required for k-wrapping later
        fft_shape_target = min(fft_shape_model_im, fft_shape_obs_psf) + 1
        object.__setattr__(self, "fft_shape_target", fft_shape_target)

    def __call__(self, kimages, key=None):

        model_kim, model_kpsf, obs_kpsf = kimages
        kimage_final = model_kim / model_kpsf * obs_kpsf
        
        kimage_final_wrap = jax.vmap(wrap_hermitian_x, in_axes=(0, None, None, None, None, None, None))(
                            kimage_final,
                            -self.fft_shape_target//2,
                            -self.fft_shape_target//2,
                            -self.fft_shape_target//2+1,
                            -self.fft_shape_target//2,
                            self.fft_shape_target-1,
                            self.fft_shape_target-1
        )

        kimage_final_wrap = kimage_final_wrap[:, :-1, :]
    
        kimg_shift = jnp.fft.ifftshift(kimage_final_wrap, axes=(-2,))

        real_image_arr = jnp.fft.fftshift(
            jnp.fft.irfft2(kimg_shift, 
                           [self.fft_shape_target-1, self.fft_shape_target-1], (-2, -1)), (-2, -1)
        )

        img_trimed = _trim(real_image_arr, [real_image_arr.shape[0], self.real_shape_target[-2], self.real_shape_target[-1]])

        return img_trimed


import operator
from functools import reduce

import jax.numpy as jnp

from . import Scenery
from . import measure
from .bbox import Box
from .morphology import ArrayMorphology, GaussianMorphology
from .observation import Observation
from .spectrum import ArraySpectrum


# function to calculate values of perimeter pixels
def _get_edge_pixels(img, box):
    box_slices = box.slices
    box_start = box.start
    box_stop = box.stop
    edge = [
        img[:, box_slices[0], box_start[1]],
        img[:, box_slices[0], box_stop[1] - 1],
        img[:, box_start[0], box_slices[1]],
        img[:, box_stop[0] - 1, box_slices[1]],
    ]
    return jnp.concatenate(edge, axis=1)


def make_bbox(obs, center_pix, min_size=11, delta_size=3, min_snr=20, min_corr=0.99):
    """
    Make a bounding box for source at center.

    This method finds small box around the center so that the edge flux has a minimum SNR,
    the color of the edge pixels remains highly correlated to the center pixel color,
    and the flux in all channels is lower than with the last acceptable box size.
    This should effectively isolate the source against the noise background and neighboring objects.

    Parameters
    ----------
    obs: `~scarlet2.Observation`
    center_pix: tuple
        source enter, in pixel coordinates
    min_size: int
        smallest box size
    delta_size: int
        incremental increase of box size
    min_snr: float
        minimum SNR of edge pixels (aggregated over all observation channel) to allow increase of box size
    min_corr: float
        minimum correlation coefficient between center and edge color to allow increase of box size

    Returns
    -------
    ~scarlet2.Box
    """
    assert isinstance(obs, Observation)
    assert obs.weights is not None, "Observation weights are required"

    peak_spectrum = pixel_spectrum(obs, center_pix, correct_psf=True, return_array=True)
    last_spectrum = peak_spectrum.copy()
    box2d = Box((min_size, min_size))
    if not obs.frame.bbox.spatial.contains(center_pix):
        raise ValueError(f"Pixel coordinate expected, got {center_pix}")
    box2d.set_center(center_pix.astype(int))

    # increase box size until SNR is below threshold or spectrum changes significantly
    while max(box2d.shape) < max(obs.frame.bbox.spatial.shape):
        edge_pixels = _get_edge_pixels(obs.data, box2d)
        edge_spectrum = jnp.mean(edge_pixels, axis=-1)
        edge_spectrum /= jnp.sqrt(jnp.dot(edge_spectrum, edge_spectrum))

        weight_edge_pixels = _get_edge_pixels(obs.weights, box2d)
        snr_edge_pixels = edge_pixels * jnp.sqrt(weight_edge_pixels)
        valid_edge_pixel = weight_edge_pixels > 0
        mean_snr = jnp.sum(jnp.sum(snr_edge_pixels, axis=-1) / jnp.sum(valid_edge_pixel, axis=-1))

        if mean_snr < min_snr:
            break

        spec_corr = jnp.dot(edge_spectrum, peak_spectrum) / \
                    jnp.sqrt(jnp.dot(peak_spectrum, peak_spectrum)) / jnp.sqrt(jnp.dot(edge_spectrum, edge_spectrum))

        if spec_corr < min_corr or jnp.any(edge_spectrum > last_spectrum):
            if min(box2d.shape) > min_size:
                box2d = box2d.shrink(delta_size)
            break

        box2d = box2d.grow(delta_size)

    box = obs.frame.bbox[0] @ box2d
    return box


def compact_morphology(min_value=1e-6, return_array=False):
    """
    Create image of the point source morphology model, i.e. the most compact source possible.

    Parameters
    ----------
    min_value: float
        Minimum pixel value (needed for positively constrained morphologies)
    return_array: bool
        Whether to return the array or an ArrayMorphology

    Returns
    -------
    2D jnp.array (normalized to (0,1)) or ArrayMorphology

    """
    try:
        frame = Scenery.scene.frame
    except AttributeError:
        print("Compact morphology can only be created within the context of a Scene")
        print("Use 'with Scene(frame) as scene: Source(...)'")
        raise
    if frame.psf is None:
        raise AttributeError("Compact morphology can only be create with a PSF in the model frame")

    morph = frame.psf.morphology()
    morph = jnp.maximum(morph, min_value)
    if return_array:
        return morph
    return ArrayMorphology(morph)


def standardized_moments(
        obs,
        center,
        footprint=None,
        bbox=None,
):
    """Create image of a Gaussian from the 2nd moments of the observation in the region of the bounding box.

    Parameters
    ----------
    obs : `~scarlet2.Observation`
    center : tuple
        central pixel of the source
    bbox: `~scarlet2.BBox`
        box to cut out source from observation, in pixel coordinates
    min_value: float
        minimum pixel value (useful to set to > 0 for positivity constraints)
    return_array: bool
        Whether to return arrays or ArraySpectrum and ArrayMorphology

    Returns
    -------
    jnp.ndarrays (for spectrum and morphology) or ArraySpectrum, ArrayMorphology
    """
    assert isinstance(obs, Observation)
    # construct box from footprint
    if bbox is None:
        assert footprint is not None
        bbox = Box.from_data(footprint)
    # construct footprint as step function inside bbox
    if footprint is None:
        assert bbox is not None
        footprint = jnp.zeros(obs.frame.bbox.spatial.shape)
        footprint = footprint.at[bbox.spatial.slices].set(1)

    # cutout image and footprint
    cutout_img = obs.data[bbox.slices]
    cutout_fp = footprint[bbox.spatial.slices]
    center -= jnp.array(bbox.spatial.origin)

    try:
        frame = Scenery.scene.frame
    except AttributeError:
        print("Adaptive morphology can only be created within the context of a Scene")
        print("Use 'with Scene(frame) as scene: Source(...)'")
        raise
    if frame.psf is None:
        raise AttributeError("Adaptive morphology can only be created with a PSF in the model frame")

    # getting moment measures:
    # 1) get convolved moments
    g = measure.moments(cutout_img, center=center, weight=cutout_fp[None, :, :], N=2)
    # 2) adjust moments for model frame
    g.transfer(obs.frame.wcs, frame.wcs)
    # 3) deconvolve from PSF (actually: difference kernel between obs PSF and model frame PSF)
    if hasattr(obs, "_dp"):
        p = obs._dp
    else:
        # moments of difference kernel between the model PSF and the observed PSF
        p = measure.moments(obs.frame.psf(), N=2)
        p.transfer(obs.frame.wcs, frame.wcs)
        p0 = measure.moments(frame.psf(), N=2)
        p.deconvolve(p0)
        # store in obs for repeated use
        object.__setattr__(obs, "_dp", p)
    # 3) deconvolve from difference kernel
    g.deconvolve(p)
    return g


def from_gaussian_moments(obs, center, min_size=11, delta_size=3, min_snr=20, min_corr=0.99, min_value=1e-6,
                          return_array=False):
    # TODO: implement with source footprints given for each observation

    # get moments from all channels in all observations
    if not isinstance(obs, (list, tuple)):
        observations = (obs,)
    else:
        observations = obs
    centers = [obs_.frame.get_pixel(center) for obs_ in observations]
    boxes = [make_bbox(obs_, center_, min_size=min_size, delta_size=delta_size, min_snr=min_snr, min_corr=min_corr) for
             obs_, center_ in zip(observations, centers)]
    moments = [standardized_moments(obs_, center_, bbox=bbox_) for obs_, center_, bbox_ in
               zip(observations, centers, boxes)]

    # flat lists of spectra, sorted as model frame channels
    spectra = jnp.concatenate([g[0, 0] for g in moments])
    channels = reduce(operator.add, [obs_.frame.channels for obs_ in observations])
    spectrum = sort_spectra(spectra, channels)

    # average over all channels
    g = moments[0]
    for key in g.keys():
        g[key] = jnp.concatenate([g[key] for g in moments])  # combine all observations
        g[key] = jnp.median(g[key])  # this is not SNR weighted nor consistent aross different moments, but works(?)

    # create morphology and evaluate at center
    morph = GaussianMorphology.from_moments(g)
    morph = morph()
    spectrum /= morph.sum()
    morph = jnp.maximum(morph, min_value)

    if return_array:
        return spectrum, morph
    return ArraySpectrum(spectrum), ArrayMorphology(morph)


# initialise the spectrum
def pixel_spectrum(obs, pos, correct_psf=False, return_array=False):
    """Get the spectrum at a given position in the observation(s).

    Yields the spectrum of a single-pixel source with flux 1 in every channel,
    concatenated for all observations.

    If `correct_psf`, it homogenizes the PSFs of the observations, which yields the
    correct spectrum for a flux=1 point source.

    If `model` is set, it reads of the value of the model at `sky_coord` and yields the
    spectrum for that model.

    Parameters
    ----------
    obs: `~scarlet2.Observation` or lists of observations
        Observation(s) to extract pixel SED from
    pos: tuple
        Position in the observation. Needs to be in sky coordinates if multiple observations have different locations
        or pixel scales.
    correct_psf: bool
        Whether PSF shape variations in the observations should be corrected
    return_array: bool
        Whether to return the array or an ArraySpectrum

    Returns
    -------
    spectrum: `~jnp.array`, ArraySpectrum, or list thereof if given list of observations
    """

    # for multiple observations, get spectrum from each observation and then combine channels in order of model frame
    if isinstance(obs, (list, tuple)):

        # flat lists of spectra and channels in order of observations
        spectra = jnp.concatenate(
            [pixel_spectrum(obs_, pos, correct_psf=correct_psf, return_array=True) for obs_ in obs])
        channels = reduce(operator.add, [obs_.frame.channels for obs_ in obs])
        spectrum = sort_spectra(spectra, channels)

        if return_array:
            return spectrum
        return ArraySpectrum(spectrum)

    assert isinstance(obs, Observation)

    pixel = obs.frame.get_pixel(pos).astype(int)

    if not obs.frame.bbox.spatial.contains(pixel):
        raise ValueError(f"Pixel coordinate expected, got {pixel}")

    spectrum = obs.data[:, pixel[0], pixel[1]].copy()

    if correct_psf and obs.frame.psf is not None:

        try:
            frame = Scenery.scene.frame
        except AttributeError:
            print("Adaptive morphology can only be created within the context of a Scene")
            print("Use 'with Scene(frame) as scene: Source(...)'")
            raise
        if frame.psf is None:
            raise AttributeError("Adaptive morphology can only be create with a PSF in the model frame")

        # correct spectrum for PSF-induced change in peak pixel intensity
        psf_model = obs.frame.psf()
        psf_peak = psf_model.max(axis=(-2, -1))

        psf0_model = frame.psf()
        psf0_peak = psf0_model.max(axis=(-2, -1))

        spectrum /= psf_peak / psf0_peak

    if jnp.any(spectrum <= 0):
        # If the flux in all channels is  <=0,
        # the new sed will be filled with NaN values,
        # which will cause the code to crash later
        msg = f"Zero or negative spectrum {spectrum} at {pos}"
        if jnp.all(spectrum <= 0):
            print("Zero or negative spectrum in all channels: Setting spectrum to 1")
            spectrum = jnp.ones_like(spectrum)
        print(msg)

    if return_array:
        return spectrum
    return ArraySpectrum(spectrum)


def sort_spectra(spectra, channels):
    try:
        frame = Scenery.scene.frame
    except AttributeError:
        print("Multi-observation initialization can only be created within the context of a Scene")
        print("Use 'with Scene(frame) as scene: ...")
        raise

    spectrum = []
    for channel in frame.channels:
        try:
            idx = channels.index(channel)
            spectrum.append(spectra[idx])
        except ValueError:
            msg = f"Channel '{channel}' not found in observations. Setting amplitude to 0."
            print(msg)
            spectrum.append(0)
    spectrum = jnp.array(spectrum)
    return spectrum



import os
import h5py
import numpy as np
import jax
import jax.numpy as jnp
from numpyro.distributions import constraints
import matplotlib.pyplot as plt

from scarlet2 import *
from utils import import_scarlet_test_data

import_scarlet_test_data()
from scarlet_test_data import data_path
from scarlet2.io import model_to_h5, model_from_h5


def test_save_output():
    file = jnp.load(os.path.join(data_path, "hsc_cosmos_35.npz"))
    data = jnp.asarray(file["images"])
    centers = [(src["y"], src["x"]) for src in file["catalog"]]  # Note: y/x convention!
    weights = jnp.asarray(1 / file["variance"])
    psf = jnp.asarray(file["psfs"])

    frame_psf = GaussianPSF(0.7)
    model_frame = Frame(Box(data.shape), psf=frame_psf)
    obs = Observation(data, weights, psf=ArrayPSF(jnp.asarray(psf))).match(model_frame)

    from functools import partial

    spec_step = partial(relative_step, factor=0.05)

    with Scene(model_frame) as scene:

        for center in centers:
            center = jnp.array(center)
            try:
                spectrum, morph = init.from_gaussian_moments(obs, center, min_corr=0.99)
            except ValueError:
                spectrum = init.pixel_spectrum(obs, center)
                morph = init.compact_morphology()

            Source(center, spectrum, morph)

    # fitting
    parameters = scene.make_parameters()
    for i in range(len(scene.sources)):
        parameters += Parameter(
            scene.sources[i].spectrum.data,
            name=f"spectrum:{i}",
            constraint=constraints.positive,
            stepsize=spec_step,
        )
        parameters += Parameter(
            scene.sources[i].morphology.data,
            name=f"morph:{i}",
            constraint=constraints.positive,
            stepsize=0.1,
        )

    maxiter = 200
    scene.set_spectra_to_match(obs, parameters)
    scene_ = scene.fit(obs, parameters, max_iter=maxiter, progress_bar=False)

    # save the output
    ID = 1
    filename = "demo_io"
    path = "stored_models"
    model_to_h5(filename, scene_, ID, path=path, overwrite=True)

    # demo that it works to add models to a single file
    ID = 2
    filename = "demo_io"
    path = "stored_models"
    model_to_h5(filename, scene_, ID, path=path, overwrite=True)

    # load files and show keys 
    with h5py.File(f"{path}/{filename}.h5", "r") as f:
        print(f.keys())

    # print the output
    print(f"Output saved to {path}/{filename}.h5")
    # print the storage size 
    print(f"Storage size: {os.path.getsize(f'{path}/{filename}.h5')/1e6:.4f} MB")
    # load the output and plot the sources
    scene_loaded = model_from_h5(filename, ID, path=path)
    print("Output loaded from h5 file")

    # compare scenes 
    saved = jax.tree_util.tree_leaves(scene_)
    loaded = jax.tree_util.tree_leaves(scene_loaded)
    status = True
    for leaf_saved, leaf_loaded in zip(saved, loaded):
        if (leaf_saved != leaf_loaded).all():
            status = False
        
    print(f"saved == loaded: {status}")
    assert status == True, "Loaded leaves not identical to original"

if __name__ == "__main__":
    test_save_output()



import equinox as eqx
import jax.numpy as jnp

from . import Scenery
from .module import Module


class Spectrum(Module):

    @property
    def shape(self):
        raise NotImplementedError


class ArraySpectrum(Spectrum):
    data: jnp.array

    def __init__(self, data):
        self.data = data

    def __call__(self):
        return self.data

    @property
    def shape(self):
        return self.data.shape


class StaticArraySpectrum(Spectrum):
    data: jnp.array
    channelindex: list = eqx.field(static=True)

    def __init__(self, data, filters):
        try:
            frame = Scenery.scene.frame
        except AttributeError:
            print("Source can only be created within the context of a Scene")
            print("Use 'with Scene(frame) as scene: Source(...)'")
            raise

        self.data = data
        self.channelindex = jnp.array([filters.index(c[0]) for c in frame.channels])

    def __call__(self):
        return self.data[self.channelindex]

    @property
    def shape(self):
        return len(self.channelindex),


class TransientArraySpectrum(Spectrum):
    data: jnp.array
    _epochmultiplier: jnp.array = eqx.field(static=True)

    def __init__(self, data, epochs):
        try:
            frame = Scenery.scene.frame
        except AttributeError:
            print("Source can only be created within the context of a Scene")
            print("Use 'with Scene(frame) as scene: Source(...)'")
            raise
        self.data = data
        self._epochmultiplier = jnp.array([1.0 if c in epochs else 0.0 for c in frame.channels])

    def __call__(self):
        return jnp.multiply(self.data, self._epochmultiplier)

    @property
    def shape(self):
        return self.data.shape



import h5py
import os
import numpy as np
import jax.numpy as jnp
from scarlet2 import *
import pickle
import jax


def model_to_h5(filename, scene, ID, path="", overwrite=False):
    """
    Save the model output to a single HDF5 file.
    NOTE: This is not a pure function hence cannot be
    utalized within a JAX JIT compilation.

    Inputs
    ------
    filename : str
    model : scarlet2.scene instance
    ID : int
    overwrite : bool, optional

    Returns
    -------
    None
    """
    # create directory if it does not exist
    if not os.path.exists(path):
        os.makedirs(path)

    # first serialize the model into a pytree
    model_group = str(ID)
    save_h5_path = os.path.join(path, filename)

    f = h5py.File(f"{save_h5_path}.h5", "a")
    # create a group for the scene
    if model_group in f.keys():
        if overwrite:
            del f[model_group]
        else:
            raise ValueError(
                "ID already exists. Set overwrite=True to overwrite the ID."
            )

    # save the binary to HDF5
    group = f.create_group(model_group)
    model = pickle.dumps(scene)
    group.attrs["model"] = np.void(model)
    f.close()


def model_from_h5(filename, ID, path=""):
    """
    Load the model output from a single HDF5 file.

    Inputs
    ------
    filename : str
    ID : int

    Returns
    -------
    scene : scarlet2.scene instance
    """

    filename = os.path.join(path, filename)
    f = h5py.File(f"{filename}.h5", "r")
    model_group = str(ID)
    if model_group not in f.keys():
        raise ValueError(f"ID {ID} not found in the file.")

    group = f.get(model_group)
    out = group.attrs["model"]
    binary_blob = out.tobytes()
    scene = pickle.loads(binary_blob)
    f.close()

    return scene



from setuptools import setup

setup(
    name='scarlet2',
    version='0.2.0',    
    description='Scarlet, all new and shiny',
    url='https://github.com/pmelchior/scarlet2',
    author='Peter Melchior',
    author_email='melchior@astro.princeton.edu',
    license='MIT',
    packages=['scarlet2'],
    install_requires=[
        'equinox',
        'jax',
        'astropy',
        'numpy',
        'matplotlib',
        'varname'
    ],
)
